// file micrec.c - MMC/SD card USB audio recorder

//#define puts ERROR_DO_NOT_USE_PUTS
#define puts(a);

#include "micrec.h"

/// \todo: make audio int unable to overwrite RIFF struct at start of file

#define MIN_FILE_SIZE 50000
#define USE_44K
#define DEBUG_LEVEL 0
#define USE_HC      /* Support SD-HC cards. (42 words) */
#define EXTEND_DIRECTORY
//#define MMC_USE_MULTIWRITE


#ifndef VCC
#define VCC
#endif
#ifdef VCC

u_int16 bufferSpace;
u_int16 bufWi;
u_int16 bufRi;
u_int16 bufferMinimum = 65535;
u_int32 originalDacInt;
u_int32 not_originalUsbInt;
u_int32 originalGpio0Int;

 /* Microphone Buffer Ptr */
/* Use USB out-packet-buffer 6 for storing sample data */
s_int16 *micBufStart = USB_SEND_MEM + (6 * 64);
s_int16 *micBufEnd = USB_SEND_MEM + (6 * 64) + 32;
s_int16 *micBufPtr = USB_SEND_MEM + (6 * 64);
volatile u_int16 edgeTime = 0;
volatile u_int16 edgeTime2 = 0;
//u_int16 mmcDoMultiWrite = 0;

// Amount of memory to use for file write buffering, excluding extra 45 words
// This must be multiple of 256 and 45 free words must follow

u_int32 currentBlock;
u_int32 currentBlockFill;
u_int16 interfaceStopped = 0;
u_int16 samplesToSave = 0;
#endif

#define DT_LANGUAGES 0
#define DT_VENDOR 1
#define DT_MODEL 2
#define DT_SERIAL 3
#define DT_DEVICE 4
#define DT_CONFIGURATION 5

#define VENDOR_NAME_LENGTH 7
const u_int16 myVendorNameStr[] = {
  ((VENDOR_NAME_LENGTH * 2 + 2)<<8) | 0x03,
  'v'<<8,'l'<<8,'s'<<8,'i'<<8,'.'<<8,'f'<<8,'i'<<8};

#define MODEL_NAME_LENGTH 26
const u_int16 myModelNameStr[] = {
  ((MODEL_NAME_LENGTH * 2 + 2)<<8) | 0x03,
  'U'<<8,
  'S'<<8,
  'B'<<8,
  'R'<<8,
  'e'<<8,
  'c'<<8,
  'o'<<8,
  'r'<<8,
  'd'<<8,
  'e'<<8,
  'r'<<8,
  ' '<<8,
  'i'<<8,
  'n'<<8,
  ' '<<8,
  'R'<<8,
  'e'<<8,
  'c'<<8,
  'o'<<8,
  'r'<<8,
  'd'<<8,
  ' '<<8,
  'M'<<8,
  'o'<<8,
  'd'<<8,
  'e'<<8
};


 // This is the new Device Descriptor. See the USB specification! 
const u_int16  myDeviceDescriptor [] = "\p"
  "\x12" // Length
  "\x01" // Type (Device Descriptor)
  "\x10" // LO(bcd USB Specification version) (.10)
  "\x01" // HI(bcd USB Specification version) (1.)
  "\x00" // Device Class (will be specified in configuration descriptor)
  "\x00" // Device Subclass
  "\x00" // Device Protocol
  "\x40" // Endpoint 0 size (64 bytes)

  "\xfb" // LO(Vendor ID) (0x19fb=VLSI Solution Oy)
  "\x19" // HI(Vendor ID)

  "\xb0" // LO(Product ID) (0xeee0 - 0xeeef : VLSI Customer Testing)
  "\xee" // HI(Product ID) (customers can request an ID from VLSI)

  "\x00" // LO(Release Number)
  "\x00" // HI(Release Number)
  "\x01" // Index of Vendor (manufacturer) name string
  "\x02" // Index of Product (model) name string (most shown by windows)
  "\x00" // Index of serial number string
  "\x01" // Number of configurations (1)
;

#define CONFIG_DESC_SIZE 174

u_int16 myConfigurationDescriptor[] = "\p"
  
  // ______Configuration_descriptor____ at offset 0
  "\x09"   // Length: 9 bytes // (9)
  "\x02"   // Descriptor type: Configuration // (2)
  "\xae"    // LO(TotalLength) (174)
  "\x00"    // HI(TotalLength) (0)
  "\x03"    // Number of Interfaces (3)
  "\x01"    // bConfigurationValue (1)
  "\x00"    // iConfiguration (0)
  "\x80"    // Attributes (128)
  "\x32"    // Max. Power 100mA (50)
  
// ---------- AUDIO CONTROL INTERFACE --------------

// ______Interface_descriptor____ at offset 9
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x00"   // Interface number  // (0)
  "\x00"   // Alternate setting  // (0)
  "\x00"   // Number of endpoints  // (0)
  "\x01"   // Interface Class: Audio // (1)
  "\x01"   // Interface Subclass: Audio Control // (1)
  "\x00"   // Interface Protocol  // (0)
  "\x00"   // String index  // (0)
  
// ______Audio Control Class Interface_descriptor____ at offset 18
  "\x0a"   // Length: 10 bytes // (10)
  "\x24"   // Descriptor type: Audio Control Class Interface // (36)
  "\x01"   // Subtype Header // (1)
  "\x00"    // (0)
  "\x01"    // (1)
  "\x34"    // (51) size 30 + 21 = 51 + 1 = 52
  "\x00"    // 
  "\x02"    // Number of interfaces
  "\x01"    // (1) audio interface 1
  "\x02"    // (2) audio interface 2
// ______Audio Control Class Interface_descriptor____ at offset 27
  "\x0C"   // Length: 12 bytes // (12)
  "\x24"   // Descriptor type: Audio Control Class Interface // (36)
  "\x02"   // Subtype Input Terminal // (2)
  "\x01"    // (1) Terminal 1
  "\x01"    // (1)
  "\x01"    // (1) 0101 usb streaming
  "\x00"    // (0) assoc. terminal 
  "\x02"    // (2) 2 channels
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)
// ______Audio Control Class Interface_descriptor____ at offset 39
  "\x09"   // Length: 9 bytes // (9)
  "\x24"   // Descriptor type: Audio Control Class Interface // (36)
  "\x03"   // Subtype Output Terminal // (3)
  "\x02"    // (2) Terminal 2
  "\x04"    // (4)
  "\x03"    // (3) 0304 headphone audio etc
  "\x00"    // (0)
  "\x01"    // (1) input is terminal 1 usb audio
  "\x00"    // (0) no string
// ______Audio Control Class Interface_descriptor____ 48
  "\x0C"   // Length: 12 bytes // (12)
  "\x24"   // Descriptor type: Audio Control Class Interface // (36)
  "\x02"   // Subtype Input Terminal // (2)
  "\x03"    // (3) Terminal 3
  "\x01"    // (1)
  "\x02"    // (2) 0x01,0x02 = 0x0201 = Microphone
  "\x00"    // (0)
  "\x01"    // (1) 1 channel
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0) no string
// ______Audio Control Class Interface_descriptor____ 60
  "\x09"   // Length: 9 bytes // (9)
  "\x24"   // Descriptor type: Audio Control Class Interface // (36)
  "\x03"   // Subtype Output Terminal // (3)
  "\x04"    // (4) Terminal 4
  "\x01"    // (4)
  "\x01"    // (3) 0101 usb streaming
  "\x00"    // (0)
  "\x03"    // (3) input is terminal 3 microphone
  "\x00"    // (0) no string


// ---------- AUDIO STREAMING INTERFACE (SPK) ------------

// ______Interface_descriptor____ at offset 69
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x01"   // Interface number  // (1)
  "\x00"   // Alternate setting  // (0)
  "\x00"   // Number of endpoints  // (0)
  "\x01"   // Interface Class: Audio // (1)
  "\x02"   // Interface Subclass: Audio Streaming // (2)
  "\x00"   // Interface Protocol  // (0)
  "\x00"   // String index  // (0)
// ______Interface_descriptor____ at offset 78
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x01"   // Interface number  // (1)
  "\x01"   // Alternate setting  // (1)
  "\x01"   // Number of endpoints  // (1)
  "\x01"   // Interface Class: Audio // (1)
  "\x02"   // Interface Subclass: Audio Streaming // (2)
  "\x00"   // Interface Protocol  // (0)
  "\x00"   // String index  // (0)
// ______Audio Streaming Class Interface_descriptor____ at offset 87
  "\x07"   // Length: 7 bytes // (7)
  "\x24"   // Descriptor type: Audio Streaming Class Interface // (36)
  "\x01"   // Subtype  // (1) GENERAL
  "\x01"    // (1) INPUT TERMNAL 1
  "\x0C"    // (12) Delay 12 frames
  "\x01"    // (1) 
  "\x00"    // (0)  PCM
// ______Audio Streaming Class Interface_descriptor____ at offset 94
  "\x0B"   // Length: 11 bytes // (11)
  "\x24"   // Descriptor type: Audio Streaming Class Interface // (36)
  "\x02"   // Subtype  // (2) FORMAT
  "\x01"    // (1) FORMAT TYPE 1
  "\x02"    // (2) CHANNELS 2
  "\x02"    // (2) SUBFRAMESIZE 2
  "\x10"    // (16) Used bits: 16
  "\x01"    // (1) Frequencies 1
  "\x44"    // (68) 
  "\xAC"    // (172) 
  "\x00"    // (0)     0x00AC44 = 44100
// ______Endpoint_descriptor____ at offset 105
  "\x09"   // Length: 9 bytes // (9)
  "\x05"   // Descriptor type: Endpoint // (5)
  "\x01"   // Endpoint Address  // (1)
  "\x01"   // Attributes.TransferType Isochronous(1) // (9)
  "\xFF"   // EP Size LSB // (255)
  "\x00"   // EP Size MSB (total 1023 bytes) // (3)
  "\x01"   // Polling Interval ms // (1)
  "\x00"   // Refresh // (0)
  "\x00"   // Sync Address // (0)
// ______Class Endpoint_descriptor____ at offset 114
  "\x07"   // Length: 7 bytes // (7)
  "\x25"   // Descriptor type: Class Endpoint // (37)
  "\x01"   // Subtype  // (1) GENERAL
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)
  

// ---------- AUDIO STREAMING INTERFACE (MIC) ------------

// ______Interface_descriptor____ at offset 121
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x02"   // Interface number  // (2)
  "\x00"   // Alternate setting  // (0)
  "\x00"   // Number of endpoints  // (0)
  "\x01"   // Interface Class: Audio // (1)
  "\x02"   // Interface Subclass: Audio Streaming // (2)
  "\x00"   // Interface Protocol  // (0)
  "\x00"   // String index  // (0)
// ______Interface_descriptor____ at offset 130
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x02"   // Interface number  // (2)
  "\x01"   // Alternate setting  // (1)
  "\x01"   // Number of endpoints  // (1)
  "\x01"   // Interface Class: Audio // (1)
  "\x02"   // Interface Subclass: Audio Streaming // (2)
  "\x00"   // Interface Protocol  // (0)
  "\x00"   // String index  // (0)
// ______Audio Streaming Class Interface_descriptor____ at offset 139
  "\x07"   // Length: 7 bytes // (7)
  "\x24"   // Descriptor type: Audio Streaming Class Interface // (36)
  "\x01"   // Subtype  // (1) GENERAL
  "\x04"    // (1) OUTPUT TERMNAL 4 (USB)
  "\x0C"    // (12) Delay 12 frames
  "\x01"    // (1) 
  "\x00"    // (0)  PCM
// ______Audio Streaming Class Interface_descriptor____ at offset 146
  "\x0B"   // Length: 11 bytes // (11)
  "\x24"   // Descriptor type: Audio Streaming Class Interface // (36)
  "\x02"   // Subtype  // (2) FORMAT
  "\x01"    // (1) FORMAT TYPE 1
  "\x01"    // (1) CHANNELS 1
  "\x02"    // (2) SUBFRAMESIZE 2
  "\x10"    // (16) Used bits: 16
  "\x01"    // (1) Frequencies 1
  "\x22"    // ( ) 
  "\x56"    // ( ) 
  "\x00"    // (0)     0x001F40=8K  0x002B11=11025 0x005622=22050
// ______Endpoint_descriptor____ at offset 157
  "\x09"   // Length: 9 bytes // (9)
  "\x05"   // Descriptor type: Endpoint // (5)
  "\x81"   // Endpoint Address  // (1) IN
  "\x01"   // Attributes.TransferType Isochronous(1)
  "\xFF"   // EP Size LSB // (255)
  "\x00"   // EP Size MSB (total 1023 bytes) // (3)
  "\x01"   // Polling Interval ms // (1)
  "\x00"   // Refresh // (0)
  "\x00"   // Sync Address // (0)
// ______Class Endpoint_descriptor____ at offset 166
  "\x07"   // Length: 7 bytes // (7)
  "\x25"   // Descriptor type: Class Endpoint // (37)
  "\x01"   // Subtype  // (1) GENERAL
  "\x00"    // (0) 
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)

  // end of descriptor
;



// When a USB setup packet is received, install our descriptors 
// and then proceed to the ROM function RealDecodeSetupPacket.
void RealInitUSBDescriptors(u_int16 initDescriptors);
void MyInitUSBDescriptors(u_int16 initDescriptors){
  RealInitUSBDescriptors(2); // ROM set descriptors for audio
  USB.descriptorTable[DT_VENDOR] = myVendorNameStr;
  USB.descriptorTable[DT_MODEL]  = myModelNameStr;
  USB.descriptorTable[DT_DEVICE] = myDeviceDescriptor;
  USB.descriptorTable[DT_CONFIGURATION] = myConfigurationDescriptor;
  USB.configurationDescriptorSize = CONFIG_DESC_SIZE;
}

#if 0
static char hex[] = "0123456789ABCDEF";
void puthex(u_int16 d) {
    register int i;
    char mem[6];
    for (i=0;i<4;i++) {
	mem[i] = hex[(d>>12)&15];
	d <<= 4;
    }
    mem[4] = ' ';
    mem[5] = '\0';
    fputs(mem, stdout);
}
#endif



#if 0
void PutBuffer(s_int16 a) {
  if (samplesToSave) {
    a = SwapWord(a);
    if (bufWi < 0x1800) {
      *(u_int16 *)bufWi = (a);
    } else {
      *(__y u_int16 *)(bufWi - 0x1800 + 0x1C00/*yOffset*/) = (a);
    }
    if (++bufWi >= BUFFER_SIZE) {
      bufWi = 0;
    }
    bufferSpace -= 1;
  }
}
#endif


#ifdef VCC
void StartADC(){
  //Save original int vector
  originalDacInt = ReadIRam(0x20+INTV_DAC);
#if USE_USB_INTERRUPT
  originalUsbInt = ReadIRam(0x20+INTV_USB);
#endif
  originalGpio0Int = ReadIRam(0x20+INTV_GPIO0);
  PERIP(INT_ENABLE) = 0; // Disable All Interrupts
  //Set new int vectors
  SetInterruptVector_GPIO0();
  SetInterruptVector_DAC();
#if USE_USB_INTERRUPT
  WriteIRam (0x20+INTV_USB, ReadIRam((u_int16)InterruptStub0));
#endif

  //Mic input GPIOs
  PERIP(GPIO0_CLEAR_MASK) = (1<<MIC_RESET_BIT); // output level is low
//PERIP(GPIO0_MODE) &= ~((1<<MIC_RESET_BIT) | (1<<MIC_INPUT_BIT)); //is GPIO
//  PERIP(GPIO0_MODE) = 0xb9f;
  PERIP(GPIO0_MODE) = 0x009f;
//PERIP(GPIO0_DDR) |= REC_LED;  
//PERIP(GPIO0_DDR) &= ~(1<<MIC_INPUT_BIT);; //is input
//  PERIP(GPIO0_DDR) = 0x3440 | REC_LED;
  PERIP(GPIO0_DDR) = 0x7E40;
//PERIP(GPIO0_INT_FALL) |= (1 << MIC_INPUT_BIT); //Falling edge int
  PERIP(GPIO0_INT_FALL) = 0x0020;

  //  puthex(PERIP(GPIO0_MODE));
  //  puthex(PERIP(GPIO0_DDR));
  //  puthex(PERIP(GPIO0_INT_FALL));
  //  puts("=perip");

  //Configure Timer1 for ramp measurement
  PERIP(TIMER_T1L) = 30000; //374=16 kHz //2176 ~is about 11025
  PERIP(TIMER_T1H) = 0x0;
  PERIP(TIMER_T1CNTL) = 1000; //Current Value Low
  PERIP(TIMER_T1CNTH) = 0; //Current Value High
  PERIP(TIMER_CONFIG) = 0; //Timer Clock is 48MHz
  PERIP(TIMER_ENABLE) |= (1 << 1); //Enable timer 1

  // Enable Interrupts
  Enable();
  PERIP(INT_ENABLEH) = 0; 
  PERIP(INT_ENABLEL) = INTF_GPIO0 | INTF_RX | INTF_DAC;

}




void StopADC(){

  Disable();	
  PERIP(INT_ENABLEH) = 0; 
  PERIP(INT_ENABLEL) = INTF_RX | INTF_DAC;
  PERIP(TIMER_ENABLE) &= ~(1 << 1); //Disable timer 1
  WriteIRam (0x20+INTV_DAC, originalDacInt);
#if USE_USB_INTERRUPT
  WriteIRam (0x20+INTV_USB, originalUsbInt);
#endif
  WriteIRam (0x20+INTV_GPIO0, originalGpio0Int);
  Enable();

}
#endif //cc


//s_int16 Sat16 (register __a s_int32 n);


#ifdef VCC
void InterruptHandler_DAC(void){ 
  static s_int16 currentSample;
  static u_int16 mode = GOING_UP;

  
  if (mode == GOING_UP) {

    PERIP(TIMER_CONFIG) = 0; // 48 MHz timer stepping rate; max@44.1 = 1088 steps
    mode = RESETTING;
    PERIP(GPIO0_ODATA) &= ~(1<<MIC_RESET_BIT); //output state is low
    PERIP(GPIO0_DDR) |= (1<<MIC_RESET_BIT); //is output
    
    // If edgeTime is sane (set at GPIO0 interrupt), use as sample
    if (edgeTime > (MIC_TIMER_START-1040) && edgeTime < (MIC_TIMER_START-40)) {
      currentSample = ((s_int16)((edgeTime)-MIC_OFFSET) * (MIC_GAIN));
    }    
    *micBufPtr++ = SwapWord(currentSample);
    if (micBufPtr > micBufEnd) micBufPtr = micBufStart;

    // EP1 IN is isochronous. Make it so. Always.
    PERIP(USB_EP_ST1) |= USB_STF_IN_ENABLE | USB_STF_IN_ISO;
    
    // If PC wants mic data, put it to xmit buffer
    if (PERIP(USB_EP_ST1)& USB_STF_IN_EMPTY){
      register s_int16 *i;
      register s_int16 *p;
      register u_int16 n;
      p = (s_int16 *) USB_SEND_MEM + (5 * 64); // use send buffer 3
      i = micBufStart;
      n = 0;
      while (i < micBufPtr){
	*p++ = *i++;
	n++;
      } 
      micBufPtr = micBufStart;
      PERIP(USB_EP_SEND1) = 0x8000 | (5 << 10) |(n*2);
    }

           
  }else{

    mode = GOING_UP;

    PERIP(GPIO0_DDR) &= ~(1<<MIC_RESET_BIT); //is input (HiZ)
    PERIP(TIMER_T1CNTL) = MIC_TIMER_START; //Current Value Low


#define DO_SATURATION
    // Store filedata
    //{
    if (samplesToSave) {
#ifdef DO_SATURATION      
      register s_int32 a;
#else
      register s_int16 a;
#endif
      a = audioPtr.rd[0];
      a += audioPtr.rd[1];
      a += currentSample;
#ifdef DO_SATURATION
      // 18 words w/s_int32
      //      if (a>32767) a = 32767;
      //      if (a<-32768) a = -32768;
      //a = Sat16(a);
#endif      
      {
	register s_int16 b = SwapWord(Sat16(a));
	if (bufWi < 0x1800) {
	  *(u_int16 *)bufWi = (b);
	} else {
	  *(__y u_int16 *)(bufWi + 0x0400/*yOffset-0x1800*/) = (b);
	}
      }
      
      if (++bufWi >= BUFFER_SIZE) {
	bufWi = 0;
      }
      bufferSpace -= 1;
      
    }    
    
    // VU meter LEDs (16 words)
    PERIP(GPIO1_CLEAR_MASK) = (LED1 | LED2);
    if (currentSample > 1000) {
      PERIP(GPIO1_SET_MASK) = LED1;
    }
    if (currentSample > 8000) {
      PERIP(GPIO1_SET_MASK) = LED2;
    }



#if 0    
    /* Handle timeCount here, because timer interrupt is disabled.
    USBWantsSuspend() does not work without timeCount. NOTE:SUSPEND DOES NOT WORK*/
    /*22050Hz*/
    {
      static int cnt = 0;
      if (++cnt > 22050/TIMER_TICKS) {
	cnt = 0;
	timeCount++;
      }
    }
#endif

  }

  
  //For debugging: put current sample to audio buffer
  audioPtr.rd[0] = currentSample;
  

}

#endif //cc





#if 0
void MyDecodeSetupPacket(){
  USB.EPReady[0] = 1; //EP0 shalt be ready now.
  if (((s_int16)USB.pkt.payload[0] >= 0 ) &&
      ((USB.pkt.payload[USB_SP_REQUEST] & 0xff) ==
       USB_REQUEST_SET_INTERFACE) &&
      (USB.pkt.payload[USB_SP_VALUE] == 0))
    {
      puts("interfaceStop");
      interfaceStopped = 1;
    };
  RealDecodeSetupPacket();
}
#endif


void SpiExec(u_int16 programOffset){
  ResetIRAMVectors();
  currentKeyMap = sixKeyMap; //in ROM
  PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
  PERIP(SPI0_FSYNC) = 0;
  PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (12-1);
  PERIP(INT_ENABLEL) = 0;
  PERIP(GPIO1_MODE) = 0x3f;	
  SpiLoad(programOffset,SPIADDR_EXTRABYTES);	    	  
}



#ifdef VCC
void main() {

  //puts("micre.c");
  InitAudio(); /* goto 3.0x..4.0x */
  PERIP(INT_ENABLEL) = INTF_RX | INTF_TIM0;
  PERIP(INT_ENABLEH) = INTF_DAC;
  samplesToSave = 0;

  originalDacInt = ReadIRam(0x20+INTV_DAC);
  SetInterruptVector_DAC();

  StartADC();
  Enable();

  voltages[voltCoreUSB] = 29;
  voltages[voltIoUSB] = 31;
  PowerSetVoltages(&voltages[voltCoreUSB]);

  PERIP(GPIO1_ODATA) |=  LED1|LED2;      /* POWER led on */
  PERIP(GPIO1_DDR)   |=  LED1|LED2; /* SI and SO to outputs */
  PERIP(GPIO1_MODE)  &= ~(LED1|LED2); /* SI and SO to GPIO */
  PERIP(GPIO0_ODATA) &= ~(1 << 10); // NFCE low

  LED_ON(UR3_LED4);


  LoadCheck(NULL, 1); /* 4.0x clock required! */
  SetRate(44100U); /* Our USB-audio requires 44100Hz */
  
  map = FsMapMmcCreate(NULL, 0); /* Create MMC Mapper */
  FsInitializeMmc(50);  

  
  bufferSpace = BUFFER_SIZE;

  FatPrepareForSaving();
  if (!USBIsAttached) {
    // if USB is not attached, start immediately recording from microphone.
    samplesToSave = 1;
  }


  //SetHookFunction((u_int16)DecodeSetupPacket, MyDecodeSetupPacket); 
  SetHookFunction((u_int16)InitUSBDescriptors, MyInitUSBDescriptors);  

  InitUSB(2);





  while (1) { 
    
    USBHandler();  
    
    if (USBWantsSuspend()) {
      //puts("UsbWantsSuspend");
      //interfaceStopped = 2;
    }
    
    { // Use last sent SET_INTERFACE value to set recording on/off
      static u_int16 lastInterfaceSetting = 0;
      if (lastInterfaceSetting != USB.interfaces){
	lastInterfaceSetting = USB.interfaces; //USB.interfaces holds last SET_INTERFACE setting
	//puthex(lastInterfaceSetting);puts("=interfaces");
	if (lastInterfaceSetting == 0x0101){
	  samplesToSave = 1;
	} else if (lastInterfaceSetting == 0x0001){
	  samplesToSave = 0;
	  interfaceStopped = 1;
	}
      }
    }




    if (mmc.state == mmcWriteWait) {
      register u_int16 c;
      if (c=SpiSendReceiveMmc(0xff00U, 8) == 0x00ffU) {
	mmc.state = mmcOk; /*write finished*/
	LED2_OFF;
      }
    } else {
      
      if (bufferSpace < (BUFFER_SIZE - 255)) {
	if (bufRi < 0x1800) {
	  memcpy(minifatBuffer, (u_int16 *)bufRi, 256);
	}else{
	  memcpyYX(minifatBuffer, (__y u_int16 *)(bufRi-0x1800+0x1c00), 256);
	}	
	bufferSpace += 256;
	REC_LED_ON;
	FsMapMmcWrite(NULL /*no wait*/, currentBlock, 1, minifatBuffer);
	mmc.fileSize += 512;
	if (mmc.fileSize / 512 >= mmc.largestFreeSize) {
	  //End of disk space (end of largest fragment)
	  //puts("End Of Disk");
	  interfaceStopped = 2;	  
	}	  
	currentBlock++;	  

	REC_LED_OFF;
	bufRi += 256;
	if (bufRi >= BUFFER_SIZE) {
	  bufRi = 0;
	}	
      } 

      if ((ReadGPIO() & (KEY_5))){ //Pressed key 5 (PLAY)
	interfaceStopped = 2; //save file and exit
      }

      if (PERIP(SCI_STATUS) & SCISTF_REGU_POWERBUT) { //power button
	interfaceStopped = 2; //save file and exit
      }




      if (ReadGPIO() & KEY_1){
	samplesToSave = 1;
      }

      {
	static u_int16 myTimeCount;
	if (((++myTimeCount >> 6) & 0x0f) == 0) {
	  REC_LED_ON;
	} else {
	  REC_LED_OFF;
	}
      }
      //puthex(timeCount);


      if (interfaceStopped){
	samplesToSave = 0;
	if (mmc.fileSize > MIN_FILE_SIZE) { //some arbitrary non-zero min file size
	  //puts("Write File");
	  //puthex(bufferMinimum);puts("=bufferMinimum");
	  bufferMinimum = 65535;	  
	  FsFatMakeFile();
	  FatPrepareForSaving();
	}
	if (interfaceStopped == 2){ //USB Detach or end of disk
	  //puts("Exit");
	  Disable();
	  PERIP(GPIO1_MODE) = 0x3f;
	  puts("ToPcPlayer");
	  SpiExec(PC_PLAYER_MODE);

	}
	interfaceStopped = 0;
      }
    }
  }
}
#endif //cc

#ifdef VCC

struct FsMapper *FsMapMmcCreate(struct FsPhysical *physical,
				u_int16 cacheSize);
u_int16 FsMapMmcRead(struct FsMapper *map, u_int32 firstBlock,
		     u_int16 blocks, u_int16 *data);
u_int16 FsMapMmcWrite(struct FsMapper *map, u_int32 firstBlock,
		      u_int16 blocks, u_int16 *data);

const struct FsMapper mmcMapper = {
    0x010c, /*version*/
    256,    /*blocksize in words*/
    0,      /*blocks -- read from CSD*/
    0,      /*cacheBlocks*/
    FsMapMmcCreate,
    FsMapFlNullOk,//RamMapperDelete,
    FsMapMmcRead,
    FsMapMmcWrite,
    NULL,//FsMapFlNullOk,//RamMapperFree,
    FsMapFlNullOk,//RamMapperFlush,
    NULL /* no physical */
};





#if 0
struct FsMapper *FsMapMmcCreate(struct FsPhysical *physical, u_int16 cacheSz);
#else
struct FsMapper *FsMapMmcCreate(struct FsPhysical *physical, u_int16 cacheSz) {

  //Configure MMC pins
  //Already done:
  //PERIP(GPIO0_MODE) &= ~(MMC_MISO|MMC_CLK|MMC_MOSI|MMC_XCS);
  //Already done:
  //PERIP(GPIO0_DDR) = (PERIP(GPIO0_DDR) & ~(MMC_MISO))
  //	| (MMC_CLK|MMC_MOSI|MMC_XCS);

    PERIP(GPIO0_ODATA) = (PERIP(GPIO0_ODATA) & ~(MMC_CLK|MMC_MOSI))
	| (MMC_XCS | GPIO0_CS1);
    
    memsetY(&mmc, 0, sizeof(mmc));
    return &mmcMapper;
}
#endif



u_int16 FsMapMmcRead(struct FsMapper *map, u_int32 firstBlock,
		     u_int16 blocks, u_int16 *data) {
    register u_int16 bl = 0;
    register u_int32 fB = firstBlock & 0x00ffffffUL; /*remove sign extension: 4G -> 8BG limit*/
    do {
	if (MyReadDiskSector(data, fB))
	    break; /* probably MMC detached */
	data += 256;
	fB++;
	bl++;
    } while (bl < blocks);
    return bl;
}


#endif //VCC



#ifdef VCC
s_int16 FsInitializeMmc(s_int16 tries) {
    register s_int16 i, cmd;
    mmc.state = mmcNA;
    mmc.blocks = 0;
    mmc.errors = 0;


 tryagain:
    IdleHook();

    mmc.hcShift = 9;
    if (tries-- <= 0) {
	return ++mmc.errors;
    }

    for (i=0; i<512; i++) {
	SpiSendClocks();
    }

    /* MMC Init, command 0x40 should return 0x01 if all is ok. */
    i = MmcCommand(MMC_GO_IDLE_STATE/*CMD0*/|0x40,0);
    if (i != 1) {
	BusyWait10();
	goto tryagain;//continue; /* No valid idle response */
    }
    cmd = MMC_SEND_OP_COND|0x40;
#ifdef USE_HC
    /*CMD8 is mandatory before ACMD41
      for hosts compliant to phys. spec. 2.00 */
    i = MmcCommand(MMC_SEND_IF_COND/*CMD8*/|0x40,
		   0x00000122/*2.7-3.6V*/); /*note: 0x22 for the right CRC!*/
    if (i == 1) {
	/* MMC answers: 0x05 illegal command,
	   v2.0 SD(HC-SD) answers: 0x01 */
	/* Should we read the whole R7 response? */
	cmd = 0x40|41; /* ACMD41 - SD_SEND_OP_COND */
    }
#endif /*USE_HC*/
    /* MMC Wake-up call, set to Not Idle (mmc returns 0x00)*/
    i = 0; /* i is 1 when entered .. but does not make the code shorter*/
    while (1) {
	register int c;
#ifdef USE_HC
	if (cmd == (0x40|41)) {
	    c = MmcCommand(0x40|55/*CMD55*/,0);
	}
#endif
	c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/,
#ifdef USE_HC
		       /* Support HC for SD, AND for MMC! */
		       //i ? 0x40100000UL : 0x00000000UL
		       0x40100000UL
#else
		       0
#endif
	    );
	if (c == 0) {
	    break;
	}
	if (++i >= 25000 /*Large value (12000) required for MicroSD*/
	    || c != 1) {
	    goto tryagain; /* Not able to power up mmc */
	}
    }
#ifdef USE_HC
    i = MmcCommand(MMC_READ_OCR/*CMD58*/|0x40, 0);
    if (//cmd == (0x40|41) &&
	i == 0) {
	if (SpiSendReceiveMmc(-1, 16) & (1<<(30-16))) {
	    /* OCR[30]:CCS - card capacity select */
	    /* HighCapacity! */
	    mmc.hcShift = 0;
	}
    }
#endif /*USE_HC*/

    if (MmcCommand(MMC_SEND_CSD/*CMD9*/|0x40, 0) == 0) {
	register s_int16 *p = (s_int16 *)minifatBuffer;
	register int t = 640;
	while (SpiSendReceiveMmc(0xff00, 8) == 0xff) {
	    if (t-- == 0) {
		goto tryagain;
	    }
	}
	for (i=0; i<8; i++) {
	    *p++ = SpiSendReceiveMmc(-1, 16);
/*
            00 00 11 11 222 2 3 3334444555566667777
  64M  MMC: 8C 0E 01 2A 03F 9 8 3D3F6D981E10A4040DF
            CSD  NSAC   CCC  flg       MUL+E
	      TAAC  SPEE   RBL  C_SIZE(>>2=FDB)
	                        33 333333333344 444 444 444 444 445
	                        00 111101001111 110 110 110 110 011 000000111
				   C_SIZE                       MULT
  256M MMC+ 90 5E 00 2A 1F5 9 8 3D3EDB683FF9640001F=CSD (Kingston)
  128M SD:  00 26 00 32 175 9 8 1E9F6DACF809640008B=CSD (DANE-ELEC)
  4G SD-HC: 40 7F 0F 32 5B5 9 8 0001E44 7F801640004B
            00 00 11 11 222 2 3 3334444 555566667777
 */
	}
	if ((minifatBuffer[0] & 0xf000) == 0x4000) {
	    /* v2.0 in 512kB resolution */
	  //puts("v2");
	    mmc.blocks = 
		(((((u_int32)minifatBuffer[3]<<16) | minifatBuffer[4])
		 & 0x3fffff) + 1) << 10;
	} else {
	    /* v1.0 */
	    register s_int32 c_size = (((minifatBuffer[3] & 0x03ff)<<2) |
				       ((minifatBuffer[4]>>14) & 3)) + 1;
	    register s_int16 c_mult = (((minifatBuffer[4] & 3)<<1) |
				       ((minifatBuffer[5]>>15) & 1));
	    //puts("v1");
	    c_size <<= 2 + c_mult + (minifatBuffer[2] & 15);
	    mmc.blocks = c_size >> 9;
	}
    }
    /* All OK return */
    //mmc.errors = 0;
    mmc.state = mmcOk;
    map->blocks = mmc.blocks;
    return 0;//mmc.errors;
}

#endif //cc

#ifdef VCC

u_int16 FsMapMmcWrite(struct FsMapper *map, u_int32 firstBlock,
		      u_int16 blocks, u_int16 *data) {
    register u_int16 bl = 0;
    register u_int32 fB = firstBlock & 0x00ffffffUL; /*remove sign extension: 4G -> 8BG limit*/
//    while (bl < blocks) {
    do {
	register u_int16 c;
	if (mmc.state != mmcOk || mmc.errors) {
	    break;
	}
#ifdef USE_HC
	  c = MmcCommand(MMC_WRITE_BLOCK|0x40, fB<<mmc.hcShift);
#else
	  c = MmcCommand(MMC_WRITE_BLOCK|0x40, fB<<9);
#endif

	//wait for BUSY token, if you get 0x01(idle), it's an ERROR!
	{
	  register int __d0 t = 0;
	  while (c != 0x00) {
	    /* timeout */
	    if (/*++t < 0 || */c == 0x01) {
	      mmc.errors++;
	      goto out;
	    }
	    c = SpiSendReceiveMmc(0xff00, 8);
	  }
	}

	  SpiSendReceiveMmc(0xfffe, 16);
#if 0
	for (c=0; c<256; c++) {
	    SpiSendReceiveMmc(*data++, 16);
	}
	SpiSendReceiveMmc(0xffff, 16); /* send CRC */
#else
	for (c=0; c<257; c++) { /*one extra for dummy CRC*/
	    SpiSendReceiveMmc(*data++, 16);
	}
#endif
	if (map) {
	    /* wait until MMC write finished */
	    do {
		/*TODO: timeout?*/
	    } while ( SpiSendReceiveMmc(0xff00U, 8) != 0x00ffU);
	} else {
	    mmc.state = mmcWriteWait;
	    /* MUST not use SpiSendClocks():
	       some MMC/SD's don't support interrupted XCS. */
	    return 0; /*Can write only one sector this way.*/
	}
	fB++;
	bl++;
    } while (bl < blocks);
 out:
    SpiSendClocks();
    return bl;
}
#endif




#ifdef VCC
u_int32 sectorsRead = 0;
auto u_int16 MyReadDiskSector(register __i0 u_int16 *buffer,
			      register __reg_a u_int32 sector) {
    register s_int16 i;
    register u_int16 t;

    sectorsRead++;
    if (mmc.state == mmcNA || mmc.errors) {
	cs.cancel = 2;
	return -1;
    }

#ifdef USE_HC
    i = MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(sector<<mmc.hcShift));
#else
    i = MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(sector<<9));
#endif
    t = 65535;
    do {
	i = SpiSendReceiveMmc(0xff00,8);
    } while (i == 0xff && --t != 0);

    if (i != 0xfe) {
	memset(buffer, 0, 256);
	if (i > 15 /*unknown error code*/) {
	    mmc.errors++;
	} else {
	    /* data retrieval failed:
	       D4-D7 = 0
	       D3 = out of range
	       D2 = Card ECC Failed
	       D1 = CC Error
	       D0 = Error
	     */
	}
	return 1;
    }
    for (i=0; i<512/2; i++) {
	*buffer++ = SpiSendReceiveMmc(0xffff,16);
    }
    SpiSendReceiveMmc(0xffff,16); /* discard crc */

    /* generate some extra SPI clock edges to finish up the command */
    SpiSendClocks();
    return 0; /* All OK return */
}


#endif //cc


#ifdef VCC
int FsFatMakeFile(void) {
    {
      register s_int32 t = mmc.fileSize - 8;
      

      FsInitializeMmc(50);

      ReadDiskSector(minifatBuffer,
		     minifatInfo.currentSector =
		     mmc.largestFreeStart);
      minifatBuffer[2] = SwapWord(t);
      minifatBuffer[3] = SwapWord(t>>16);
      t -= 0x24;
      minifatBuffer[20] = SwapWord(t);
      minifatBuffer[21] = SwapWord(t>>16);
      /* We don't care if it fails or not. */
      FsMapMmcWrite(map, mmc.largestFreeStart, 1,
		    minifatBuffer);
    }
    
    {
      register u_int32 cluster =
	(mmc.largestFreeStart - minifatInfo.dataStart)
	/ minifatInfo.fatSectorsPerCluster;
      
      
      if (mmc.dirLine < 0)
	return 1;
      
      /* allocate sectors */
      FsFatAllocate2(mmc.largestFreeStart, mmc.fileSize / 512, 0x0fffffffUL, 0);
      ReadDiskSector(minifatBuffer, minifatInfo.currentSector = mmc.dirBlock);
      
      /*4B4F4546494C45315241570000C62E000000700033000000FFFFDA09FFFFFFFF
	name -----------------
	attr 11               --
	ntres 12                --
	crt time 13 14 16         tttimedate
	last access date 18                 ----
	cluster hi 20                           ----
	write time 22 24                            timedate
	cluster low 26                                      ----
	filesize 28                                             --------
      */
      {
	register u_int16 *p;
	p = minifatBuffer + mmc.dirLine/2;
	memset(p, 0, 32/2);
	memcpyYX(p, mmc.fileName, 12/2);
	p += 10/2;
	//*p |= attrib;
	p += (20-10)/2;
	*p = SwapWord(cluster>>16);//Cluster HI
	p += (26-20)/2;
	*p = SwapWord(cluster); //Cluster LO
	p += (28-26)/2;
	p[0] = SwapWord(mmc.fileSize);
	p[1] = SwapWord(mmc.fileSize>>16);
      }
      if (FsMapMmcWrite(map, mmc.dirBlock, 1, minifatBuffer) != 1) {
      }
      return 0;
    }
}
#endif //VCC




#ifdef VCC
void FatPrepareForSaving(void) {
    u_int16 name[12/2];


#if 1
    static const u_int16 waveHeader[] = {
      /* 22050Hz 1-ch 16-bit signed WAV */
      0x5249,0x4646,0x2400,0x0000,0x5741,0x5645,0x666d,0x7420,
      0x1000,0x0000,0x0100,0x0100, //channels
      0x2256,0x0000, //samplerate
      0x44ac,0x0000, //datarate
      0x0200, //recordsize
      0x1000,0x6461,0x7461,0xffff,0xffff,0x0000,0x0000,
      0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000
      /* 10 zero-samples inserted to align to 32-word boundary */
    };       
#endif

#if 0
    static const u_int16 waveHeader[] = {
      /* 8000Hz 1-ch 16-bit signed WAV */
      0x5249,0x4646,0x2400,0x0000,0x5741,0x5645,0x666d,0x7420,
      0x1000,0x0000,0x0100,0x0100,0x401f,0x0000,0x803e,0x0000,
      0x0200,0x1000,0x6461,0x7461,0xffff,0xffff,0x0000,0x0000,
      0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000
      /* 10 zero-samples inserted to align to 32-word boundary */
    };       
#endif


#if 0
    static const u_int16 waveHeader[] = {
      /* 44100Hz 2-ch 16-bit signed WAV */
      0x5249,0x4646,0xffff,0x0000,0x5741,0x5645,0x666d,0x7420,
      0x1000,0x0000,0x0100,0x0200,0x44ac,0x0000,0x10b1,0x0200,
      0x0400,0x1000,0x6461,0x7461,0xffff,0xffff,0x0000,0x0000,
      0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000
      /* 10 zero-samples inserted to align to 32-word boundary */
    };
#endif


#if 0
    static const u_int16 waveHeader[] = {
      /* 44100Hz 1-ch 16-bit signed WAV */
      0x5249,0x4646,0xffff,0x0000,0x5741,0x5645,0x666d,0x7420,
      0x1000,0x0000,0x0100,0x0100,0x44ac,0x0000,0x8858,0x0100,
      0x0400,0x1000,0x6461,0x7461,0xffff,0xffff,0x0000,0x0000,
      0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000
      /* 10 zero-samples inserted to align to 32-word boundary */
    };
#endif


#if 0
    static const u_int16 waveHeader[] = {
      /* 16000Hz 2-ch 16-bit signed WAV */
      0x5249,0x4646,0xffff,0x0000,0x5741,0x5645,0x666d,0x7420,
      0x1000,0x0000,0x0100,0x0200,0x803e,0x0000,0x00fa,0x0000,
      0x0400,0x1000,0x6461,0x7461,0xffff,0xffff,0x0000,0x0000,
      0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000
      /* 10 zero-samples inserted to align to 32-word boundary */
    };
#endif
    

    mmc.dirLine = -1;
    mmc.largestFreeSize = 0;
    //mmc.smallerFreeSize = 0;
    mmc.fileSize = 0;
    if (InitFileSystem()) {
	//mmc.dirBlock = 1234;
	return;
    }
    //mmc.fileSize = 0;
    if (!mmc.fileName[3]) {
	memcpyXY(mmc.fileName, "\pVOICE_00WAV", sizeof(mmc.fileName));
    }
    /* Must find free area first to be able to extend directory! */
    FatIterateOverFreeSectors(FsGetFreeSectors, NULL);
    do {
	if ((++mmc.fileName[3] & 0xff) > '9') {
	    if (
		(mmc.fileName[3] += 0x100-10) > (('9'<<8)+0xff)
		) {
		mmc.fileName[3] = 0x3030; /* "00" */
		/* avoid '[' '\' ']' and '@' */
		if ((++mmc.fileName[2] & 0xff) > 'Z')
		    mmc.fileName[2] = (('E'<<8) + 'A');
	    }
	}
	memcpyYX(name, mmc.fileName, sizeof(name));
    } while (FatFindDirEntry(name) == FFDE_NAMEINVALID);

  currentBlock = mmc.largestFreeStart;

#if 0
  yBufPtr = yBufStart;
  yBufRdPtr = yBufStart;
  bufferSpace = Y_BUFFER_SIZE;
  memcpyXY(yBufPtr, waveHeader, sizeof(waveHeader));
  yBufPtr += 32;
#else
  memcpy(0, waveHeader, sizeof(waveHeader));
  bufWi = 32;
  bufRi = 0;
  bufferSpace = BUFFER_SIZE - 32;
#endif
}


#endif //CC


#ifdef VCC
 auto s_int16 FatFindDirEntry(const u_int16 *packedName /*8.3 name*/) {
    register __i1 __y struct FRAGMENT *curFragment;
    __y struct FRAGMENT * __y nextFragment;
    register __b u_int32 currentSector;
    register __i3 int i;

    /* Start at the start of root directory. */
    if (minifatInfo.IS_FAT_32) {
	nextFragment = FatFragmentList(&minifatFragments[0], 2);
    } else /*if (minifatInfo.FilSysType != 0x3231)*/ {
	minifatFragments[0].start = minifatInfo.rootStart | LAST_FRAGMENT;
	minifatFragments[0].size  = ((s_int16)minifatInfo.BPB_RootEntCnt >> 4);
	nextFragment = &minifatFragments[1];
    }/* else {
	return FFDE_FAT12;	//FAT12 not supported
    }*/

 again:
    mmc.dirLine = -1;
    curFragment = &minifatFragments[0];
    minifatInfo.parentDir = currentSector = curFragment->start & 0x7fffffffUL;
    i = 0;
    goto getfirst;
    while (1) {
	register __d1 u_int16 fn = FatGetByte(i+0); /*first char of filename*/
	if (fn == 0xe5 || fn == 0) {/* 0xe5 / 0x00 deleted or unused entry */
	    /* Found! */
	    if (mmc.dirLine < 0) {
		mmc.dirLine = i;
		mmc.dirBlock = currentSector;
	    }
	} else {
	    register int attr = FatGetByte(i+11/*Attr*/);
	    if ((attr & 0xde) == 0) {
		/* Regular file -- check filename */
		if ((memcmp(minifatBuffer+i/2, packedName, 11/2) |
		     ((minifatBuffer[i/2+11/2] & 0xff00) - packedName[11/2]))
		    == 0) {
		    return FFDE_NAMEINVALID;
		}
	    }
	}
	i = (i + 32) & 511;
	if (i == 0) {
	    /* if end of directory block, get the next one */
	    currentSector++;
	    if (currentSector >=
		(curFragment->start & 0x7fffffffUL) + curFragment->size) {
		if ((s_int32)curFragment->start < 0) {
		    if (mmc.dirLine < 0) {
#ifdef EXTEND_DIRECTORY
			/* Increase the directory size. */
			if (minifatInfo.IS_FAT_32) {
			    mmc.dirLine = 0;
			    if ((mmc.dirBlock = mmc.smallerFreeStart)) {
				/* another entry is not created, so leave
				   the start address un-updated. */
				//mmc.smallerFreeStart += minifatInfo.fatSectorsPerCluster;
				mmc.smallerFreeSize -= minifatInfo.fatSectorsPerCluster;
			    } else {
				mmc.dirBlock = mmc.largestFreeStart;
				mmc.largestFreeStart += minifatInfo.fatSectorsPerCluster;
				mmc.largestFreeSize -= minifatInfo.fatSectorsPerCluster;
			    }

			    /*Allocate new cluster for directory. */
			    FsFatAllocate2(mmc.dirBlock,
					   1 /*fatSectorsPerCluster*/,
					   0x0fffffffUL,
					   0);
			    /*Fix the cluster chain link. */
			    FsFatAllocate2(currentSector-1,
					   1 /*fatSectorsPerCluster*/,
					   (mmc.dirBlock-minifatInfo.dataStart)
					   / minifatInfo.fatSectorsPerCluster,
					   0);
			    /*Clear the new dirblocks */
			    memset(minifatBuffer, 0, sizeof(minifatBuffer));
			    for (i=0; i<minifatInfo.fatSectorsPerCluster; i++){
				FsMapMmcWrite(map,
					      mmc.dirBlock+(u_int16)i,
					      1,
					      minifatBuffer);
			    }
			} else
#endif
			{
			    return FFDE_NOSPACE;
			}
		    }
		    return FFDE_OK;
		}
		curFragment++;
		currentSector = curFragment->start & 0x7fffffffUL;
	    }
	getfirst:
	    ReadDiskSector(minifatBuffer,
			   minifatInfo.currentSector = currentSector);
	}
    }
}


#endif //cc

/*
  Can be used for allocating a range of clusters (sectors),
  and rewriting old links.
 */

#ifdef VCC
void FsFatAllocate2(u_int32 start, u_int32 length, u_int32 link, u_int16 free){
    register u_int32 fatSector;
    register u_int16 fatLine;
    length = (length + minifatInfo.fatSectorsPerCluster - 1)
	/ minifatInfo.fatSectorsPerCluster;
    start = (start - minifatInfo.dataStart) / minifatInfo.fatSectorsPerCluster;
    if (minifatInfo.IS_FAT_32) {
	fatSector = minifatInfo.fatStart + ((s_int32)start >> 7);
	fatLine   = ((u_int16)start << 2);
    } else {
	fatSector = minifatInfo.fatStart + ((s_int32)start >> 8);
	fatLine   = ((u_int16)start << 1);
    }
    fatLine &= 511;

    goto read;
    while (length) {
	length = length-1;
	{
	    register u_int32 t = ++start; /* points to next cluster here */
	    if (length == 0)
		t = link; /* last one: 0x0fffffffUL for end or next chain */
	    if (free)
		t = 0; /* Free it.. */
	    minifatBuffer[fatLine/2] = SwapWord(t);
	    fatLine += 2;
	    if (minifatInfo.IS_FAT_32) {
		/*fatLine has been added by 2, so use fatLine/2 again! */
		minifatBuffer[fatLine/2] = SwapWord(t>>16);
		fatLine += 2; /*another 2 makes 4 total*/
	    }
	}
	fatLine &= 511;
	if (fatLine == 0 || length == 0) {
	  FsMapMmcWrite(map, fatSector, 1, minifatBuffer);
	}
	if (fatLine != 0)
	    continue;
	fatSector++;
    read:
	ReadDiskSector(minifatBuffer,
		       minifatInfo.currentSector = fatSector);
    }
}
#endif //CC



#ifdef VCC
s_int16 FsGetFreeSectors(void *private, u_int32 sector, u_int32 numSecs) {
    if (numSecs > mmc.largestFreeSize) {
#ifdef EXTEND_DIRECTORY
	mmc.smallerFreeSize = mmc.largestFreeSize;
	mmc.largestFreeSize = numSecs;
	mmc.smallerFreeStart = mmc.largestFreeStart;
	mmc.largestFreeStart = sector;
#else
	mmc.largestFreeSize = numSecs;
	mmc.largestFreeStart = sector;
#endif
    }
    return 0;
}






#endif //VCC
