#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vs1000.h>
#include <audio.h>
#include <mappertiny.h>
#include <minifat.h>
#include <codec.h>
#include <vsNand.h>
#include <player.h>
#include <usblowlib.h>
#include "multi.h"
#include <romfont.h>
extern struct CodecServices cs;

//#define puts ERROR_DO_NOT_USE_PUTS
#define puts(a);


#define USE_WAV /* Play linear wav files using CodMiniWav from libdev1000. */

//#define USE_DEBUG
//#define DEBUG_LEVEL 3

#define USE_HC      /* Support SD-HC cards. */
#define PATCH_LBAB 
/* Removes 4G restriction from USB (SCSI).
   Also detects MMC/SD removal while attached to USB.
   (62 words) */


#ifdef PATCH_LBAB
#include <scsi.h>
extern struct SCSIVARS {
  SCSIStageEnum State;
  SCSIStatusEnum Status;
  u_int16 *DataOutBuffer;
  u_int16 DataOutSize;
  u_int16 DataBuffer[32];
  u_int16 *DataInBuffer;
  u_int16 DataInSize;
  unsigned int BlocksLeftToSend;
  unsigned int BlocksLeftToReceive;
  u_int32 CurrentDiskSector;
  u_int32 mapperNextFlushed;
  u_int16 cswPacket[7];  /*< command reply (Command Status Wrapper) data */
  u_int32 DataTransferLength; /*< what is expected by CBW */
  s_int32 Residue; /*< difference of what is actually transmitted by SCSI */
  s_int16 DataDirection;
} SCSI;
#endif/*PATCH_LBAB*/


#include <dev1000.h>

enum mmcState {
    mmcNA = 0,
    mmcOk = 1,
};
struct {
    enum mmcState state;
    s_int16 errors;
    s_int16 hcShift;
    u_int32 blocks;
} mmc;


#ifdef USE_DEBUG
static char hex[] = "0123456789ABCDEF";
void puthex(u_int16 d) {
    register int i;
    char mem[5];
    for (i=0;i<4;i++) {
	mem[i] = hex[(d>>12)&15];
	d <<= 4;
    }
    mem[4] = '\0';
    fputs(mem, stdout);
}
#endif

extern struct FsNandPhys fsNandPhys;
extern struct FsPhysical *ph;
extern struct FsMapper *map;
extern struct Codec *cod;
extern struct CodecServices cs;
extern u_int16 codecVorbis[];



enum customKeyEvent {
  ke_rec_lp = 64,
  ke_rec_sp,
  ke_stop,
};
const struct KeyMapping usbRecorderKeyMap[] = {
  {KEY_POWER, (int)ke_stop},

  {KEY_LONG_PRESS|KEY_POWER, ke_powerOff},

  {KEY_1, (int)ke_rec_lp},
  {KEY_1 | KEY_LONG_PRESS, (int)ke_rec_lp},

  {KEY_2, ke_earSpeaker},
  {KEY_2 | KEY_LONG_PRESS, ke_ff_faster},

  {KEY_3, ke_next},
  {KEY_3 | KEY_LONG_PRESS, ke_volumeUp},

  {KEY_4, ke_previous},
  {KEY_4 | KEY_LONG_PRESS, ke_volumeDown},

  {KEY_5, ke_pauseToggle},

  {KEY_LONG_PRESS|KEY_RELEASED, ke_ff_off}, 
  {0, ke_null}
};








s_int16 InitializeMmc(s_int16 tries) {
    register s_int16 i, cmd;
    mmc.state = mmcNA;
    mmc.blocks = 0;
    mmc.errors = 0;

#if DEBUG_LEVEL > 1
    puthex(clockX);
    puts(" clockX");
#endif
 tryagain:
    IdleHook();

    mmc.hcShift = 9;
    if (tries-- <= 0) {
	return ++mmc.errors;
    }

    for (i=0; i<512; i++) {
	SpiSendClocks();
    }

    /* MMC Init, command 0x40 should return 0x01 if all is ok. */
    i = MmcCommand(MMC_GO_IDLE_STATE/*CMD0*/|0x40,0);
    if (i != 1) {
#if DEBUG_LEVEL > 1
	puthex(i);
	puts(" Init");
#endif
	BusyWait10();
	goto tryagain;//continue; /* No valid idle response */
    }
    cmd = MMC_SEND_OP_COND|0x40;
#ifdef USE_HC
    /*CMD8 is mandatory before ACMD41
      for hosts compliant to phys. spec. 2.00 */
    i = MmcCommand(MMC_SEND_IF_COND/*CMD8*/|0x40,
		   0x00000122/*2.7-3.6V*/); /*note: 0x22 for the right CRC!*/
#if DEBUG_LEVEL > 2
    puthex(i);
    puts("=IF_COND");
#endif
    if (i == 1) {
	/* MMC answers: 0x05 illegal command,
	   v2.0 SD(HC-SD) answers: 0x01 */
	/* Should we read the whole R7 response? */
#if DEBUG_LEVEL > 1
	//SpiSendReceiveMmc(-1, 32);
	puthex(SpiSendReceiveMmc(-1, 16));
	puthex(SpiSendReceiveMmc(-1, 16));
	puts("=R7");
#endif
	cmd = 0x40|41; /* ACMD41 - SD_SEND_OP_COND */
    }
#endif /*USE_HC*/


#if 0
    do {
	i = MmcCommand(MMC_READ_OCR/*CMD58*/|0x40, 0);
#if DEBUG_LEVEL > 1
	puthex(i);
	puthex(SpiSendReceiveMmc(-1, 16));
	puthex(SpiSendReceiveMmc(-1, 16));
	puts("=READ_OCR");
#endif
    } while (0);
#endif

    /* MMC Wake-up call, set to Not Idle (mmc returns 0x00)*/
    i = 0; /* i is 1 when entered .. but does not make the code shorter*/
    while (1) {
	register int c;
#ifdef USE_HC
	if (cmd == (0x40|41)) {
	    c = MmcCommand(0x40|55/*CMD55*/,0);
#if DEBUG_LEVEL > 2
	    puthex(c);
	    puts("=CMD55");
#endif
	}
#endif

	c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/,
#ifdef USE_HC
		       /* Support HC for SD, AND for MMC! */
		       //i ? 0x40100000UL : 0x00000000UL
		       0x40100000UL
#else
		       0
#endif
	    );

#if DEBUG_LEVEL > 1
	puthex(c);
	puts("=ACMD41");
#endif

	if (c == 0) {
#if DEBUG_LEVEL > 1
	    puts("got 0");
#endif
	    break;
	}
	//BusyWait10();
	if (++i >= 25000 /*Large value (12000) required for MicroSD*/
	    || c != 1) {
#if 0
	    /* Retry with CMD1 */
	    if (c == 0xff && i == 2) {
		cmd = MMC_SEND_OP_COND|0x40;
		continue;
	    }
#endif

#if DEBUG_LEVEL > 1
	    puthex(c);
	    puts(" Timeout 2");
#endif
	    goto tryagain; /* Not able to power up mmc */
	}
    }
#ifdef USE_HC
    i = MmcCommand(MMC_READ_OCR/*CMD58*/|0x40, 0);
#if DEBUG_LEVEL > 1
    puthex(i);
    puts("=READ_OCR");
#endif
    if (//cmd == (0x40|41) &&
	i == 0) {
	if (SpiSendReceiveMmc(-1, 16) & (1<<(30-16))) {
	    /* OCR[30]:CCS - card capacity select */
	    /* HighCapacity! */
#if DEBUG_LEVEL > 1
	    puts("=HC");
#endif
	    mmc.hcShift = 0;
	}
	//SpiSendReceiveMmc(-1, 16);
    }
#endif /*USE_HC*/

    if (MmcCommand(MMC_SEND_CSD/*CMD9*/|0x40, 0) == 0) {
	register s_int16 *p = (s_int16 *)minifatBuffer;
	register int t = 640;
	while (SpiSendReceiveMmc(0xff00, 8) == 0xff) {
	    if (t-- == 0) {
#if DEBUG_LEVEL > 1
		puts("Timeout 3");
#endif
		goto tryagain;
	    }
	}
	for (i=0; i<8; i++) {
	    *p++ = SpiSendReceiveMmc(-1, 16);
#if DEBUG_LEVEL > 1
	    puthex(p[-1]);
#endif
/*
            00 00 11 11 222 2 3 3334444555566667777
  64M  MMC: 8C 0E 01 2A 03F 9 8 3D3F6D981E10A4040DF
            CSD  NSAC   CCC  flg       MUL+E
	      TAAC  SPEE   RBL  C_SIZE(>>2=FDB)
	                        33 333333333344 444 444 444 444 445
	                        00 111101001111 110 110 110 110 011 000000111
				   C_SIZE                       MULT
  256M MMC+ 90 5E 00 2A 1F5 9 8 3D3EDB683FF9640001F=CSD (Kingston)
  128M SD:  00 26 00 32 175 9 8 1E9F6DACF809640008B=CSD (DANE-ELEC)

  4G SD-HC: 40 7F 0F 32 5B5 9 8 0001E44 7F801640004B
            00 00 11 11 222 2 3 3334444 555566667777


 */
	}
	if ((minifatBuffer[0] & 0xf000) == 0x4000) {
	    /* v2.0 in 512kB resolution */
	  //puts("v2");
	    mmc.blocks = 
		(((((u_int32)minifatBuffer[3]<<16) | minifatBuffer[4])
		 & 0x3fffff) + 1) << 10;
	} else {
	    /* v1.0 */
	    register s_int32 c_size = (((minifatBuffer[3] & 0x03ff)<<2) |
				       ((minifatBuffer[4]>>14) & 3)) + 1;
	    register s_int16 c_mult = (((minifatBuffer[4] & 3)<<1) |
				       ((minifatBuffer[5]>>15) & 1));
	    //puts("v1");
	    c_size <<= 2 + c_mult + (minifatBuffer[2] & 15);
	    mmc.blocks = c_size >> 9;
	}
#if DEBUG_LEVEL > 1
	puts("=CSD");
	puthex(mmc.blocks>>16);
	puthex(mmc.blocks);
	puts("=mmcBlocks");
#endif
    }
#if DEBUG_LEVEL > 4
    if (MmcCommand(MMC_SEND_CID/*CMD10*/|0x40, 0) == 0) {
	register s_int16 *p = minifatBuffer;
	register int t = 3200;
	while (SpiSendReceiveMmc(0xff00, 8) == 0xff) {
	    if (t-- == 0)
		goto tryagain;
	}
	for (i=0; i<8; i++) {
	    *p++ = SpiSendReceiveMmc(-1, 16);
#if DEBUG_LEVEL > 1
	    puthex(p[-1]);
#endif
/*
       Man     Productname   serial#   date
          App             rev        res    crc7+stop
  4G:  1D 4144 0000000000 00 0000157A 0 06A E3
  64M: 02 0000 53444D3036 34 40185439 0 C77 75
       00 0011 1122223333 44 44555566 6 677 77
*/
	}
#if DEBUG_LEVEL > 1
	puts("=CID");
#endif
    }
#endif

#if 1
    /* Set Block Size of 512 bytes -- default for at least HC */
    /* Needed by MaxNova S043618ATA 2J310700 MV016Q-MMC */
    /* Must check return value! (MicroSD restart!) */
    {
	register int c;
	if ((c = MmcCommand(MMC_SET_BLOCKLEN|0x40, 512)) != 0) {
#if DEBUG_LEVEL > 1
	    puthex(c);
	    puts(" SetBlockLen failed");
#endif
	    goto tryagain;
	}
    }
#endif
    /* All OK return */
    //mmc.errors = 0;
    mmc.state = mmcOk;
    map->blocks = mmc.blocks;
    return 0;//mmc.errors;
}


auto u_int16 MyReadDiskSector(register __i0 u_int16 *buffer,
			      register __reg_a u_int32 sector) {
    register s_int16 i;
    register u_int16 t = 65535;

    if (mmc.state == mmcNA || mmc.errors) {
	cs.cancel = 1;
	return 5;
    }
#if 0 && defined(USE_DEBUG)
    puthex(sector>>16);
    puthex(sector);
    puts("=ReadDiskSector");
#endif
#ifdef USE_HC
    MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(sector<<mmc.hcShift));
#else
    MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(sector<<9));
#endif
    do {
	i = SpiSendReceiveMmc(0xff00,8);
    } while (i == 0xff && --t != 0);

    if (i != 0xfe) {
	memset(buffer, 0, 256);
	if (i > 15 /*unknown error code*/) {
	    mmc.errors++;
#if 0
	    putch('R');
#endif
	} else {
	    /* data retrieval failed:
	       D4-D7 = 0
	       D3 = out of range
	       D2 = Card ECC Failed
	       D1 = CC Error
	       D0 = Error
	     */
	}
	SpiSendClocks();
	return 1;
    }
    for (i=0; i<512/2; i++) {
	*buffer++ = SpiSendReceiveMmc(0xffff,16);
    }
    SpiSendReceiveMmc(0xffff,16); /* discard crc */

    /* generate some extra SPI clock edges to finish up the command */
    SpiSendClocks();
    SpiSendClocks();

    /* We force a call of user interface after each block even if we
       have no idle CPU. This prevents problems with key response in
       fast play mode. */
    IdleHook();
    return 0; /* All OK return */
}

struct FsMapper *FsMapMmcCreate(struct FsPhysical *physical,
				u_int16 cacheSize);
u_int16 FsMapMmcRead(struct FsMapper *map, u_int32 firstBlock,
		     u_int16 blocks, u_int16 *data);
u_int16 FsMapMmcWrite(struct FsMapper *map, u_int32 firstBlock,
		      u_int16 blocks, u_int16 *data);

const struct FsMapper mmcMapper = {
    0x010c, /*version*/
    256,    /*blocksize in words*/
    0,      /*blocks -- read from CSD*/
    0,      /*cacheBlocks*/
    FsMapMmcCreate,
    FsMapFlNullOk,//RamMapperDelete,
    FsMapMmcRead,
    FsMapMmcWrite,
    NULL,//FsMapFlNullOk,//RamMapperFree,
    FsMapFlNullOk,//RamMapperFlush,
    NULL /* no physical */
};
struct FsMapper *FsMapMmcCreate(struct FsPhysical *physical,
				u_int16 cacheSize) {
    PERIP(GPIO0_MODE) &= ~(MMC_MISO|MMC_CLK|MMC_MOSI|MMC_XCS);
    PERIP(GPIO0_DDR) = (PERIP(GPIO0_DDR) & ~(MMC_MISO))
	| (MMC_CLK|MMC_MOSI|MMC_XCS);
    PERIP(GPIO0_ODATA) = (PERIP(GPIO0_ODATA) & ~(MMC_CLK|MMC_MOSI))
	| (MMC_XCS | GPIO0_CS1); /* NFCE high */
#if DEBUG_LEVEL > 1
    puts("Configured MMC pins\n");
#endif
    memset(&mmc, 0, sizeof(mmc));
    return &mmcMapper;
}
u_int16 FsMapMmcRead(struct FsMapper *map, u_int32 firstBlock,
		     u_int16 blocks, u_int16 *data) {
    register u_int16 bl = 0;
#ifndef PATCH_LBAB /*if not patched already*/
    firstBlock &= 0x00ffffff; /*remove sign extension: 4G -> 8BG limit*/
#endif
    while (bl < blocks) {
	if (MyReadDiskSector(data, firstBlock))
	    break; /* probably MMC detached */
	data += 256;
	firstBlock++;
	bl++;
    }
    return bl;
}
u_int16 FsMapMmcWrite(struct FsMapper *map, u_int32 firstBlock,
		      u_int16 blocks, u_int16 *data) {
    u_int16 bl = 0;
#ifndef PATCH_LBAB /*if not patched already*/
    firstBlock &= 0x00ffffff; /*remove sign extension: 4G -> 8BG limit*/
#endif
    while (bl < blocks) {
	register u_int16 c;
	if (mmc.state == mmcNA || mmc.errors)
	    break;
#ifdef USE_HC
	c = MmcCommand(MMC_WRITE_BLOCK|0x40, firstBlock<<mmc.hcShift);
#else
	c = MmcCommand(MMC_WRITE_BLOCK|0x40, firstBlock<<9);
#endif
	//wait for BUSY token, if you get 0x01(idle), it's an ERROR!
	while (c != 0x00) {
	    if (c == 0x01) {
		mmc.errors++;
		goto out;
	    }
	    c = SpiSendReceiveMmc(0xff00, 8);
	    /*TODO: timeout?*/
	}
	SpiSendReceiveMmc(0xfffe, 16);
	for (c=0; c<256; c++) {
	    SpiSendReceiveMmc(*data++, 16);
	}
	SpiSendReceiveMmc(0xffff, 16); /* send dummy CRC */

	/* wait until MMC write finished */
	while ( SpiSendReceiveMmc(0xffffU, 16) != 0xffffU) {
	    /*TODO: timeout?*/
	}
	firstBlock++;
	bl++;
    }
 out:
    SpiSendClocks();
    return bl;
}



#ifdef USE_WAV
#include <codecminiwav.h>
#include <dev1000.h>
extern u_int16 codecVorbis[];
extern u_int16 ogg[];
extern u_int16 mInt[];

extern struct CodecServices cs;
extern struct Codec *cod;
enum CodecError PlayWavOrOggFile(void) {
  register enum CodecError ret = ceFormatNotFound;
  const char *eStr;
  LoadCheck(NULL, 0); /* higher clock, but 4.0x not absolutely required */
  if ((cod = CodMiniWavCreate())) {
      ret = cod->Decode(cod, &cs, &eStr);
      cod->Delete(cod);
      {
	  register int i;
	  for (i=0;i<DEFAULT_AUDIO_BUFFER_SAMPLES/(sizeof(tmpBuf)/2);i++) {
	      /* Must be memset each time, because effects like EarSpeaker are
		 performed in-place in the buffer */
	      memset(tmpBuf, 0, sizeof(tmpBuf));
	      AudioOutputSamples(tmpBuf, sizeof(tmpBuf)/2);
	  }
      }
      if (ret == ceFormatNotSupported)
	  return ceFormatNotFound; /* makes prev work with unsupported files */
      if (ret != ceFormatNotFound)
	  return ret;
      cs.Seek(&cs, 0, SEEK_SET);
  }
  return RealPlayCurrentFile();
}
#endif/*USE_WAV*/





void SpiExec(u_int16 programOffset){
  //puthex(programOffset);
  //puts("SpiExec from mmc.c");
  ResetIRAMVectors();
  currentKeyMap = sixKeyMap; //in ROM
  PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
  PERIP(SPI0_FSYNC) = 0;
  PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (12-1);
  PERIP(INT_ENABLEL) = 0;
  PERIP(GPIO1_MODE) = 0x3f;	
  SpiLoad(programOffset,SPIADDR_EXTRABYTES);	    	  
}



void USBRecorderKeyEventHandler(enum keyEvent event) {
  
  switch ((int)event) {
  case ke_rec_lp:
    SpiExec(VOICE_RECORD_MODE);
    break;
  case ke_rec_sp:
    puts("keyevent->losslessrecord");
    SpiExec(LOSSLESS_RECORD_MODE);
    break;
  case ke_stop:
    if (USBIsAttached()){
      SpiExec(PC_PLAYER_MODE);
    } else {
      player.pauseOn = 1;
    }
    break;   
    
  default:
    RealKeyEventHandler(event);
  }
  
}



u_int32 songNumberLed;

void MyUserInterfaceIdleHook(void) { /*94 words*/
    if (uiTrigger) {
	uiTrigger = 0;
	KeyScanNoUSB(0x1f);
	if (!(ReadGPIO() & 0x80)) {
	  // Record high quality
	  KeyEventHandler(ke_rec_sp);
	}


	{
	  static u_int16 counter = 0;
	  register u_int16 leds = 0;
	  if (((uiTime >> 3) % (player.currentFile + 2))<=(player.currentFile)){
	    if ((uiTime & 7) == 0) {
	      leds = LED1;
	    }
	  }
	  PERIP(GPIO1_ODATA) = (PERIP(GPIO1_ODATA) & ~(LED1 | LED2)) | leds;
	}
    }
}


auto s_int16 Sine(register __a0 u_int16 ph);
void Beep(){
  //int freq, int len) {
  register u_int16 phl = 0;
  register s_int16 len = 1024;
  register s_int16 *p = tmpBuf;
  //    puts("fuu");
  while (1) {
    p[0] = p[1] = Sine(phl)/2;//ISine(phl);
    phl += 2048;
    p += 2;
    if (p >= tmpBuf+sizeof(tmpBuf)) {
      p = tmpBuf;
      AudioOutputSamples(p, sizeof(tmpBuf)/2);
      len -= sizeof(tmpBuf)/2;
      if (len < 0)
        break;
    }
  }  
}


void main(void) {
  
  PERIP(INT_ENABLEL) = INTF_RX | INTF_TIM0;
  PERIP(INT_ENABLEH) = INTF_DAC;
  ResetIRAMVectors();
  Enable();
  
  puts("mmc.c");
  
  
#if 1 /*Perform some extra inits because we are started from SPI boot. */
  LoadCheck(NULL,1);
  BusyWait10();
  InitAudio(); /* goto 3.0x..4.0x */
  SetRate(44100);
  //Beep();
#endif
  
  puts("Beep");
  while (ReadGPIO() & KEY_5)
    ;
  
  SetHookFunction((u_int16)IdleHook, MyUserInterfaceIdleHook);
  currentKeyMap = usbRecorderKeyMap;
  SetHookFunction((u_int16)KeyEventHandler, USBRecorderKeyEventHandler);
  
#ifdef USE_WAV
  /* allow both .WAV and .OGG */
  supportedFiles = defSupportedFiles;
  /* Install the new file player function and return to player. */
  SetHookFunction((u_int16)PlayCurrentFile, PlayWavOrOggFile);
#endif
  
#if 0 /* Available only on newest (currently unreleased) developer library. */
  SetHookFunction((u_int16)OpenFile, FatFastOpenFile); /*Faster!*/
#endif
  //voltages[voltCorePlayer] = voltages[voltCoreUSB] = 29;//27;
  //voltages[voltIoUSB] = voltages[voltIoPlayer] = 27; /* 3.3V */
  voltages[voltAnaPlayer] = 30; /*3.6V for high-current MMC/SD!*/
  PowerSetVoltages(&voltages[voltCorePlayer]);
  
#if 0
  /* Set differential audio mode, i.e. invert the other channel. */
  audioPtr.leftVol = -audioPtr.rightVol;
#endif
  
  /* Set the leds after nand-boot! */
  PERIP(GPIO1_ODATA) |=  LED1|LED2;      /* POWER led on */
  PERIP(GPIO1_DDR)   |=  LED1|LED2; /* SI and SO to outputs */
  PERIP(GPIO1_MODE)  &= ~(LED1|LED2); /* SI and SO to GPIO */
  
  
  /* Replicate main loop */
  //player.volume = 17;
  player.volumeOffset = -24;
  player.pauseOn = 0;
  //player.randomOn = 0;
  keyOld = KEY_POWER;
  keyOldTime = -32767; /* ignores the first release of KEY_POWER */
  
  /*
    MMC mapper replaces FLASH mapper, so we do not need to hook
    ReadDiskSector.
    
    We must also replace the MassStorage() functionality, because it
    automatically deletes mapper, creates a read-write FLASH mapper,
    and restores read-only FLASH mapper at exit.
    
    We need to replace the main loop because we want to detect
    MMC/SD insertion and removal.
    
    We do not hook the MassStorage, because we can call our own version
    from the main loop without using the RAM hook. This saves some
    instruction words.
  */
  
  puts("create MMC");
  map = FsMapMmcCreate(NULL, 0); /* Create MMC Mapper */
  PlayerVolume();
  puts("Player");
  while (1) {
    /* If MMC is not available, try to reinitialize it. */
    if (mmc.state == mmcNA || mmc.errors) {
#ifdef USE_DEBUG
      puts("InitializeMmc(50)");
#endif
      InitializeMmc(50);
    }
    
    
    
    /* Try to init FAT. */
    if (InitFileSystem() == 0) {
      /* Restore the default suffixes. */
      minifatInfo.supportedSuffixes = supportedFiles;
      player.totalFiles = OpenFile(0xffffU);
      
      if (player.totalFiles == 0) {
	/* If no files found, output some samples.
	   This causes the idle hook to be called, which in turn
	   scans the keys and you can turn off the unit if you like.
	*/
#ifdef USE_DEBUG
	puts("no .ogg");
#endif
	goto noFSnorFiles;
      }
      
      /*
	Currently starts at the first file after a card has
	been inserted.
	
	Possible improvement would be to save play position
	to SPI EEPROM for a couple of the last-seen MMC/SD's.
	Could also save the play mode, volume, and earspeaker
	state to EEPROM.
      */
      player.nextStep = 1;
      player.nextFile = 0;
      while (1) {

	LED_ON(UR3_LED8);


	/* If random play is active, get the next random file,
	   but try to avoid playing the same song twice in a row. */
	if (player.randomOn) {
	  register s_int16 nxt;
	retoss:
	  nxt = rand() % player.totalFiles;
	  if (nxt == player.currentFile && player.totalFiles > 1)
	    goto retoss;
	  player.currentFile = nxt;
	} else {
	  player.currentFile = player.nextFile;
	}
	if (player.currentFile < 0)
	  player.currentFile += player.totalFiles;
	if (player.currentFile >= player.totalFiles)
	  player.currentFile -= player.totalFiles;
	player.nextFile = player.currentFile + 1;
	
	/* If the file can be opened, start playing it. */
	if (OpenFile(player.totalFiles-1-player.currentFile) < 0) {
	  //player.volumeOffset = -12; /*default volume offset*/
	  //PlayerVolume();
	  player.ffCount = 0;
	  cs.cancel = 0;
	  cs.goTo = -1; /* start playing from the start */
	  cs.fileSize = cs.fileLeft = minifatInfo.fileSize;
	  cs.fastForward = 1; /* reset play speed to normal */
	  
	  {
	    u_int16 n = player.currentFile;
	    u_int16 i = 0;
	    
	    //puthex(n+1);puts("=n");
	    
	    uiTime = player.currentFile*8;
	    songNumberLed = 0;
	    while (n){
	      if (n>4){
		songNumberLed |= (7 << i);
		i += 4;
		n -= 5;
	      } else {
		songNumberLed |= (1 << i);
		n--;
		i += 2;
	      }			
	    }		     		      
	  }
	  
	  {
	    register s_int16 oldStep = player.nextStep;
	    register s_int16 ret;
	    ret = PlayCurrentFile();
	    
	    /* If unsupported, keep skipping */
	    if (ret == ceFormatNotFound)
	      player.nextFile = player.currentFile + player.nextStep;
	    /* If a supported file found, restore play direction */
	    if (ret == ceOk && oldStep == -1)
	      player.nextStep = 1;
	  }
	} else {
	  player.nextFile = 0;
	}
	/* Leaves play loop when MMC changed */
	if (mmc.state == mmcNA || mmc.errors) {
	  break;
	}
      }
    } else {
      /* If not a valid FAT (perhaps because MMC/SD is not inserted),
	 just send some samples to audio buffer and try again. */
    noFSnorFiles:
      LoadCheck(&cs, 32); /* decrease or increase clock */
      memset(tmpBuf, 0, sizeof(tmpBuf));
      AudioOutputSamples(tmpBuf, sizeof(tmpBuf)/2);
      /* When no samples fit, calls the user interface
	 -- handles volume control and power-off. */
    }
  }
}
