//
// mmcmi.c - MMC/SD MSC + Microphone/Speaker Audio
// Based on TinyPlayer With Microphone Demo Project
// Panu-Kristian Poiksalo, VLSI Solution Oy 2008
//

/*
  Copyright 2008 VLSI Solution Oy. Absolutely no warranty.
  See project "TinyPlayer with Microphone" for notes
*/  

#define puts(a);
// ERROR_DO_NOT_USE_PUTS


#define REC_LED (1 << 14)

#define USE_MMC
#define DEBUG_LEVEL 0


#include <stdio.h> //Standard io
#include <stdlib.h> // VS_DSP Standard Library
#include <vs1000.h> // VS1000B register definitions
#include <vectors.h> // VS1000B vectors (interrupts and services)
#include <minifat.h> // Read Only Fat Filesystem
#include <mapper.h> // Logical Disk
#include <string.h> // memcpy etc
#include <player.h> // VS1000B default ROM player
#include <audio.h> // DAC output
#include <codec.h> // CODEC interface
#include <vsNand.h>
#include <mappertiny.h>
#include <usb.h>
#include <scsi.h>
#include "multi.h"
#include "dev1000.h"

#ifdef USE_MMC
#include "module_mmc.cc"
#else
#include "module_spiflash.cc"
#endif

#include "convert.h"
#define USE_USB_INTERRUPT 1
//#include "module_convert.c"


static char hex[] = "0123456789ABCDEF";
void puthex(u_int16 d) {
    register int i;
    char mem[6];
    for (i=0;i<4;i++) {
	mem[i] = hex[(d>>12)&15];
	d <<= 4;
    }
    mem[4] = ' ';
    mem[5] = '\0';
    fputs(mem, stdout);
}



u_int32 originalDacInt;
u_int32 originalUsbInt;
u_int32 originalGpio0Int;
u_int16 ticks;


enum customKeyEvent {
  ke_rec_lp = 64,
  ke_rec_sp,
  ke_stop,
  ke_mmcplay,
};


const struct KeyMapping usbRecorderKeyMap[] = {
  //{KEY_POWER, (int)ke_stop},

  {KEY_LONG_PRESS|KEY_POWER, ke_powerOff},

  {KEY_1, (int)ke_rec_lp},
  {KEY_1 | KEY_LONG_PRESS, (int)ke_rec_lp},

  {KEY_2, ke_earSpeaker},
  {KEY_2 | KEY_LONG_PRESS, ke_ff_faster},

  {KEY_3, ke_volumeUp},
  {KEY_3 | KEY_LONG_PRESS, ke_volumeUp},

  {KEY_4, ke_volumeDown},
  {KEY_4 | KEY_LONG_PRESS, ke_volumeDown},

  {KEY_5, (int)ke_mmcplay},
  {KEY_5 | KEY_LONG_PRESS|KEY_RELEASED, (int)ke_mmcplay},

  {KEY_LONG_PRESS|KEY_RELEASED, ke_ff_off}, 
  {0, ke_null}
};


void SpiExec(u_int16 programOffset){
  puts("SpiExec from mmcmi.c");
  PERIP(SCI_STATUS) &= ~SCISTF_USB_PULLUP_ENA; /* Take us 'out of the bus' */
  ResetIRAMVectors();
  currentKeyMap = sixKeyMap; //in ROM
  PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
  PERIP(SPI0_FSYNC) = 0;
  PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (12-1);
  PERIP(INT_ENABLEL) = 0;
  PERIP(GPIO1_MODE) = 0x3f;	
  SpiLoad(programOffset,SPIADDR_EXTRABYTES);	    	  
}


void USBRecorderKeyEventHandler(enum keyEvent event) {
  
  switch ((int)event) {
  case ke_rec_lp:
    SpiExec(VOICE_RECORD_MODE);
    break;
  case ke_rec_sp:
    SpiExec(LOSSLESS_RECORD_MODE);
    break;
  case ke_stop:    
    break;   
  case ke_mmcplay:
    puts("Play");
    SpiExec(STANDALONE_PLAYER_MODE);
    break;
  
  default:
    RealKeyEventHandler(event);
  }
  
}






 /* Microphone Buffer Ptr */
/* Use USB out-packet-buffer 6 for storing sample data */
s_int16 *micBufStart = USB_SEND_MEM + (6 * 64);
s_int16 *micBufEnd = USB_SEND_MEM + (6 * 64) + 32;
  s_int16 *micBufPtr = USB_SEND_MEM + (6 * 64);
volatile u_int16 edgeTime = 0;
volatile u_int16 edgeTime2 = 0;

void StartADC(){
  //Save original int vector
  originalDacInt = ReadIRam(0x20+INTV_DAC);
#if USE_USB_INTERRUPT
  originalUsbInt = ReadIRam(0x20+INTV_USB);
#endif
  originalGpio0Int = ReadIRam(0x20+INTV_GPIO0);

  PERIP(INT_ENABLE) = 0; // Disable All Interrupts
  //Set new int vector
  SetInterruptVector_GPIO0();
  SetInterruptVector_DAC();
#if USE_USB_INTERRUPT
  WriteIRam (0x20+INTV_USB, ReadIRam((u_int16)InterruptStub0));
#endif

  //Mic input GPIOs
  PERIP(GPIO0_CLEAR_MASK) = (1<<MIC_RESET_BIT); // output level is low
  PERIP(GPIO0_MODE) &= ~(1<<MIC_RESET_BIT); //is GPIO
  PERIP(GPIO0_DDR) &= ~(1<<MIC_INPUT_BIT);; //is input
  PERIP(GPIO0_MODE) &= ~(1<<MIC_INPUT_BIT); //is GPIO
  PERIP(GPIO0_INT_FALL) |= (1 << MIC_INPUT_BIT); //Falling edge int

  //Configure Timer1 for ramp measurement
  PERIP(TIMER_T1L) = 30000; //374=16 kHz //2176 ~is about 11025
  PERIP(TIMER_T1H) = 0x0;
  PERIP(TIMER_T1CNTL) = 1000; //Current Value Low
  PERIP(TIMER_T1CNTH) = 0; //Current Value High
  PERIP(TIMER_CONFIG) = 0; //Timer Clock is 48MHz
  PERIP(TIMER_ENABLE) |= (1 << 1); //Enable timer 1

  // Enable Interrupts
  Enable();
  PERIP(INT_ENABLEH) = 0; 
  PERIP(INT_ENABLEL) = INTF_GPIO0 | INTF_RX | INTF_DAC;

}

void StopADC(){

  Disable();	
  PERIP(INT_ENABLEH) = 0; 
  PERIP(INT_ENABLEL) = INTF_RX | INTF_DAC;
  PERIP(TIMER_ENABLE) &= ~(1 << 1); //Disable timer 1
  WriteIRam (0x20+INTV_DAC, originalDacInt);
#if USE_USB_INTERRUPT
  WriteIRam (0x20+INTV_USB, originalUsbInt);
#endif
  WriteIRam (0x20+INTV_GPIO0, originalGpio0Int);
  Enable();

}

void InterruptHandler_Timer1(void){}

void InterruptHandler_DAC(void){ 
  static u_int16 mode = GOING_UP;
  
  if (mode == GOING_UP) {
    static s_int16 currentSample;
    static s_int16 previousSample;
    

    PERIP(TIMER_CONFIG) = 0; // 48 MHz timer stepping rate; max@44.1 = 1088 steps
    mode = RESETTING;
    PERIP(GPIO0_ODATA) &= ~(1<<MIC_RESET_BIT); //output state is low
    PERIP(GPIO0_DDR) |= (1<<MIC_RESET_BIT); //is output
    
    // If edgeTime is sane (set at GPIO0 interrupt), use as sample
    previousSample = currentSample;
    if (edgeTime > MIC_TIMER_START-1040 && edgeTime < MIC_TIMER_START-40) {
      currentSample = ((s_int16)((edgeTime)-MIC_OFFSET) * (MIC_GAIN/2));
    }    
    *micBufPtr++ = SwapWord(currentSample+(previousSample/2));
    if (micBufPtr > micBufEnd) micBufPtr = micBufStart;

    // EP1 IN is isochronous. Make it so. Always.
    PERIP(USB_EP_ST1) |= USB_STF_IN_ENABLE | USB_STF_IN_ISO;
    
    // If PC wants mic data, put it to xmit buffer
    if (PERIP(USB_EP_ST1)& USB_STF_IN_EMPTY){
      register s_int16 *i;
      register s_int16 *p;
      register u_int16 n;
      p = (s_int16 *) USB_SEND_MEM + (5 * 64); // use send buffer 3
      i = micBufStart;
      n = 0;
      while (i < micBufPtr){
	*p++ = *i++;
	n++;
      } 
      micBufPtr = micBufStart;
      PERIP(USB_EP_SEND1) = 0x8000 | (5 << 10) |(n*2);
    }

           
  }else{

    //PERIP(INT_ENABLEL) &= ~INTF_USB;
    mode = GOING_UP;

    PERIP(GPIO0_DDR) &= ~(1<<MIC_RESET_BIT); //is input (HiZ)
    PERIP(TIMER_T1CNTL) = MIC_TIMER_START; //Current Value Low

    /* Handle timeCount here, because timer interrupt is disabled.
    USBWantsSuspend() does not work without timeCount. NOTE:SUSPEND DOES NOT WORK*/
    /*22050Hz*/
    {
      static int cnt = 0;
      if (++cnt > 22050/TIMER_TICKS) {
	cnt = 0;
	timeCount++;
	ticks++;
      }
    }

  }
}






#define DT_LANGUAGES 0
#define DT_VENDOR 1
#define DT_MODEL 2
#define DT_SERIAL 3
#define DT_DEVICE 4
#define DT_CONFIGURATION 5

#define VENDOR_NAME_LENGTH 7
const u_int16 myVendorNameStr[] = {
  ((VENDOR_NAME_LENGTH * 2 + 2)<<8) | 0x03,
  'v'<<8,'l'<<8,'s'<<8,'i'<<8,'.'<<8,'f'<<8,'i'<<8};

#define MODEL_NAME_LENGTH 29
const u_int16 myModelNameStr[] = {
  ((MODEL_NAME_LENGTH * 2 + 2)<<8) | 0x03,
  'U'<<8,
  'S'<<8,
  'B'<<8,
  'R'<<8,
  'e'<<8, //5
  'c'<<8,
  'o'<<8,
  'r'<<8,
  'd'<<8,
  'e'<<8, //10
  'r'<<8,
  ' '<<8,
  'i'<<8,
  'n'<<8,
  ' '<<8, //15
  'P'<<8,
  'C'<<8,
  ' '<<8,
  'P'<<8,
  'l'<<8, //20
  'a'<<8,
  'y'<<8,
  'e'<<8,
  'r'<<8,
  ' '<<8, //25
  'M'<<8,
  'o'<<8,
  'd'<<8,
  'e'<<8  //29
};



#define SERIAL_NUMBER_LENGTH 12
u_int16 mySerialNumberStr[] = {
  ((SERIAL_NUMBER_LENGTH * 2 + 2)<<8) | 0x03,
  '0'<<8, // You should put your own serial number here
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '0'<<8,
  '1'<<8, // Serial number should be unique for each unit
};

// This is the new Device Descriptor. See the USB specification! 
const u_int16  myDeviceDescriptor [] = "\p"
  "\x12" // Length
  "\x01" // Type (Device Descriptor)
  "\x10" // LO(bcd USB Specification version) (.10)
  "\x01" // HI(bcd USB Specification version) (1.)
  "\x00" // Device Class (will be specified in configuration descriptor)
  "\x00" // Device Subclass
  "\x00" // Device Protocol
  "\x40" // Endpoint 0 size (64 bytes)

  "\xfb" // LO(Vendor ID) (0x19fb=VLSI Solution Oy)
  "\x19" // HI(Vendor ID)

  "\xb1" // LO(Product ID) (0xeee0 - 0xeeef : VLSI Customer Testing)
  "\xee" // HI(Product ID) (customers can request an ID from VLSI)

  "\x00" // LO(Release Number)
  "\x00" // HI(Release Number)
  "\x01" // Index of Vendor (manufacturer) name string
  "\x02" // Index of Product (model) name string (most shown by windows)
  "\x00" // Index of serial number string
  "\x01" // Number of configurations (1)
;

#define CONFIG_DESC_SIZE 197

u_int16 myConfigurationDescriptor[] = "\p"
  
  // ______Configuration_descriptor____ at offset 0
  "\x09"   // Length: 9 bytes // (9)
  "\x02"   // Descriptor type: Configuration // (2)
  "\xc5"    // LO(TotalLength) (197)
  "\x00"    // HI(TotalLength) (0)
  "\x04"    // Number of Interfaces (4)
  "\x01"    // bConfigurationValue (1)
  "\x00"    // iConfiguration (0)
  "\x80"    // Attributes (128)
  "\x32"    // Max. Power 100mA (50)
  
// ---------- AUDIO CONTROL INTERFACE --------------

// ______Interface_descriptor____ at offset 9
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x00"   // Interface number  // (0)
  "\x00"   // Alternate setting  // (0)
  "\x00"   // Number of endpoints  // (0)
  "\x01"   // Interface Class: Audio // (1)
  "\x01"   // Interface Subclass: Audio Control // (1)
  "\x00"   // Interface Protocol  // (0)
  "\x00"   // String index  // (0)
  
// ______Audio Control Class Interface_descriptor____ at offset 18
  "\x0a"   // Length: 10 bytes // (10)
  "\x24"   // Descriptor type: Audio Control Class Interface // (36)
  "\x01"   // Subtype Header // (1)
  "\x00"    // (0)
  "\x01"    // (1)
  "\x34"    // (51) size 30 + 21 = 51 + 1 = 52
  "\x00"    // 
  "\x02"    // Number of interfaces
  "\x01"    // (1) audio interface 1
  "\x02"    // (2) audio interface 2
// ______Audio Control Class Interface_descriptor____ at offset 27
  "\x0C"   // Length: 12 bytes // (12)
  "\x24"   // Descriptor type: Audio Control Class Interface // (36)
  "\x02"   // Subtype Input Terminal // (2)
  "\x01"    // (1) Terminal 1
  "\x01"    // (1)
  "\x01"    // (1) 0101 usb streaming
  "\x00"    // (0) assoc. terminal 
  "\x02"    // (2) 2 channels
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)
// ______Audio Control Class Interface_descriptor____ at offset 39
  "\x09"   // Length: 9 bytes // (9)
  "\x24"   // Descriptor type: Audio Control Class Interface // (36)
  "\x03"   // Subtype Output Terminal // (3)
  "\x02"    // (2) Terminal 2
  "\x04"    // (4)
  "\x03"    // (3) 0304 headphone audio etc
  "\x00"    // (0)
  "\x01"    // (1) input is terminal 1 usb audio
  "\x00"    // (0) no string
// ______Audio Control Class Interface_descriptor____ 48
  "\x0C"   // Length: 12 bytes // (12)
  "\x24"   // Descriptor type: Audio Control Class Interface // (36)
  "\x02"   // Subtype Input Terminal // (2)
  "\x03"    // (3) Terminal 3
  "\x01"    // (1)
  "\x02"    // (2) 0x01,0x02 = 0x0201 = Microphone
  "\x00"    // (0)
  "\x01"    // (1) 1 channel
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0) no string
// ______Audio Control Class Interface_descriptor____ 60
  "\x09"   // Length: 9 bytes // (9)
  "\x24"   // Descriptor type: Audio Control Class Interface // (36)
  "\x03"   // Subtype Output Terminal // (3)
  "\x04"    // (4) Terminal 4
  "\x01"    // (4)
  "\x01"    // (3) 0101 usb streaming
  "\x00"    // (0)
  "\x03"    // (3) input is terminal 3 microphone
  "\x00"    // (0) no string


// ---------- AUDIO STREAMING INTERFACE (SPK) ------------

// ______Interface_descriptor____ at offset 69
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x01"   // Interface number  // (1)
  "\x00"   // Alternate setting  // (0)
  "\x00"   // Number of endpoints  // (0)
  "\x01"   // Interface Class: Audio // (1)
  "\x02"   // Interface Subclass: Audio Streaming // (2)
  "\x00"   // Interface Protocol  // (0)
  "\x00"   // String index  // (0)
// ______Interface_descriptor____ at offset 78
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x01"   // Interface number  // (1)
  "\x01"   // Alternate setting  // (1)
  "\x01"   // Number of endpoints  // (1)
  "\x01"   // Interface Class: Audio // (1)
  "\x02"   // Interface Subclass: Audio Streaming // (2)
  "\x00"   // Interface Protocol  // (0)
  "\x00"   // String index  // (0)
// ______Audio Streaming Class Interface_descriptor____ at offset 87
  "\x07"   // Length: 7 bytes // (7)
  "\x24"   // Descriptor type: Audio Streaming Class Interface // (36)
  "\x01"   // Subtype  // (1) GENERAL
  "\x01"    // (1) INPUT TERMNAL 1
  "\x0C"    // (12) Delay 12 frames
  "\x01"    // (1) 
  "\x00"    // (0)  PCM
// ______Audio Streaming Class Interface_descriptor____ at offset 94
  "\x0B"   // Length: 11 bytes // (11)
  "\x24"   // Descriptor type: Audio Streaming Class Interface // (36)
  "\x02"   // Subtype  // (2) FORMAT
  "\x01"    // (1) FORMAT TYPE 1
  "\x02"    // (2) CHANNELS 2
  "\x02"    // (2) SUBFRAMESIZE 2
  "\x10"    // (16) Used bits: 16
  "\x01"    // (1) Frequencies 1
  "\x44"    // (68) 
  "\xAC"    // (172) 
  "\x00"    // (0)     0x00AC44 = 44100
// ______Endpoint_descriptor____ at offset 105
  "\x09"   // Length: 9 bytes // (9)
  "\x05"   // Descriptor type: Endpoint // (5)
  "\x01"   // Endpoint Address  // (1)
  "\x01"   // Attributes.TransferType Isochronous(1) // (9)
  "\xFF"   // EP Size LSB // (255)
  "\x00"   // EP Size MSB (total 1023 bytes) // (3)
  "\x01"   // Polling Interval ms // (1)
  "\x00"   // Refresh // (0)
  "\x00"   // Sync Address // (0)
// ______Class Endpoint_descriptor____ at offset 114
  "\x07"   // Length: 7 bytes // (7)
  "\x25"   // Descriptor type: Class Endpoint // (37)
  "\x01"   // Subtype  // (1) GENERAL
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)
  

// ---------- AUDIO STREAMING INTERFACE (MIC) ------------

// ______Interface_descriptor____ at offset 121
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x02"   // Interface number  // (2)
  "\x00"   // Alternate setting  // (0)
  "\x00"   // Number of endpoints  // (0)
  "\x01"   // Interface Class: Audio // (1)
  "\x02"   // Interface Subclass: Audio Streaming // (2)
  "\x00"   // Interface Protocol  // (0)
  "\x00"   // String index  // (0)
// ______Interface_descriptor____ at offset 130
  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x02"   // Interface number  // (2)
  "\x01"   // Alternate setting  // (1)
  "\x01"   // Number of endpoints  // (1)
  "\x01"   // Interface Class: Audio // (1)
  "\x02"   // Interface Subclass: Audio Streaming // (2)
  "\x00"   // Interface Protocol  // (0)
  "\x00"   // String index  // (0)
// ______Audio Streaming Class Interface_descriptor____ at offset 139
  "\x07"   // Length: 7 bytes // (7)
  "\x24"   // Descriptor type: Audio Streaming Class Interface // (36)
  "\x01"   // Subtype  // (1) GENERAL
  "\x04"    // (1) OUTPUT TERMNAL 4 (USB)
  "\x0C"    // (12) Delay 12 frames
  "\x01"    // (1) 
  "\x00"    // (0)  PCM
// ______Audio Streaming Class Interface_descriptor____ at offset 146
  "\x0B"   // Length: 11 bytes // (11)
  "\x24"   // Descriptor type: Audio Streaming Class Interface // (36)
  "\x02"   // Subtype  // (2) FORMAT
  "\x01"    // (1) FORMAT TYPE 1
  "\x01"    // (1) CHANNELS 1
  "\x02"    // (2) SUBFRAMESIZE 2
  "\x10"    // (16) Used bits: 16
  "\x01"    // (1) Frequencies 1
  "\x22"    // ( ) 
  "\x56"    // ( ) 
  "\x00"    // (0)     0x001F40=8K  0x002B11=11025 0x005622=22050
// ______Endpoint_descriptor____ at offset 157
  "\x09"   // Length: 9 bytes // (9)
  "\x05"   // Descriptor type: Endpoint // (5)
  "\x81"   // Endpoint Address  // (1) IN
  "\x01"   // Attributes.TransferType Isochronous(1)
  "\xFF"   // EP Size LSB // (255)
  "\x00"   // EP Size MSB (total 1023 bytes) // (3)
  "\x01"   // Polling Interval ms // (1)
  "\x00"   // Refresh // (0)
  "\x00"   // Sync Address // (0)
// ______Class Endpoint_descriptor____ at offset 166
  "\x07"   // Length: 7 bytes // (7)
  "\x25"   // Descriptor type: Class Endpoint // (37)
  "\x01"   // Subtype  // (1) GENERAL
  "\x00"    // (0) 
  "\x00"    // (0)
  "\x00"    // (0)
  "\x00"    // (0)


// ------------ MASS STORAGE INTERFACE ------------
// ______Interface_descriptor____ at offset 9

  "\x09"   // Length: 9 bytes // (9)
  "\x04"   // Descriptor type: Interface // (4)
  "\x03"   // Interface number  // (3)
  "\x00"   // Alternate setting  // (0)
  "\x02"   // Number of endpoints  // (2)
  "\x08"   // Interface Class: Mass Storage // (8)
  "\x06"   // Interface Subclass:  // (6)
  "\x50"   // Interface Protocol  // (80)
  "\x00"   // String index  // (0)
  
  // ______Endpoint_descriptor____ at offset 18
  "\x07"   // Length: 7 bytes // (7)
  "\x05"   // Descriptor type: Endpoint // (5)
  "\x82"   // Endpoint Address  // (130)
  "\x02"   // Attributes.TransferType Bulk(2) // (2)
  "\x40"   //EP Size LSB // (64)
  "\x00"   //EP Size MSB (total 64 bytes) // (0)
  "\x10"   //Polling Interval ms // (16)
  
  // ______Endpoint_descriptor____ at offset 25
  "\x07"   // Length: 7 bytes // (7)
  "\x05"   // Descriptor type: Endpoint // (5)
  "\x03"   // Endpoint Address  // (3)
  "\x02"   // Attributes.TransferType Bulk(2) // (2)
  "\x40"   //EP Size LSB // (64)
  "\x00"   //EP Size MSB (total 64 bytes) // (0)
  "\x10"   //Polling Interval ms // (16)
  // end of descriptor
;



// When a USB setup packet is received, install our descriptors 
// and then proceed to the ROM function RealDecodeSetupPacket.
void RealInitUSBDescriptors(u_int16 initDescriptors);
void MyInitUSBDescriptors(u_int16 initDescriptors){
  RealInitUSBDescriptors(1); // ROM set descriptors for mass storage
  USB.descriptorTable[DT_VENDOR] = myVendorNameStr;
  USB.descriptorTable[DT_MODEL]  = myModelNameStr;
  USB.descriptorTable[DT_SERIAL] = mySerialNumberStr;
  USB.descriptorTable[DT_DEVICE] = myDeviceDescriptor;
  USB.descriptorTable[DT_CONFIGURATION] = myConfigurationDescriptor;
  USB.configurationDescriptorSize = CONFIG_DESC_SIZE;
}

#if 0
// this can be helpful in making a serial number
const u_int16  bHexChar16[] = { // swapped Unicode hex characters
  0x3000, 0x3100, 0x3200, 0x3300, 0x3400, 0x3500, 0x3600, 0x3700,
  0x3800, 0x3900, 0x4100, 0x4200, 0x4300, 0x4400, 0x4400, 0x4500
};My
void MakeSerialNumber(u_int32 newSerialNumber){
  u_int16 i;
  u_int32 mySerialNumber = newSerialNumber;
  // Put unique serial number to serial number descriptor
  for (i=5; i<13; i++){
    mySerialNumberStr[i]=bHexChar16[mySerialNumber>>28];
    mySerialNumber <<= 4;
  }
}
#endif

extern struct FsNandPhys fsNandPhys;
extern struct FsPhysical *ph;
extern struct FsMapper *map;
extern struct Codec *cod;
extern struct CodecServices cs;
extern u_int16 codecVorbis[];


// Get Packet from USB receive ring buffer
u_int16 GetPacket(USBPacket *packet) {
  register __c0 u_int16 st;
  register __c1 u_int16 ep;
  register __b0 u_int16 tmp;
  register __b1 u_int16 lenW;
  /*
   * There should be now some data waiting at receive buffer
   */
  st = USEX(PERIP(USB_RDPTR) + USB_RECV_MEM);
  ep = (s_int16)(st & 0x3C00) >> 10;
  st &= 0x03FF;
  packet->length = st;

  lenW = (st + 1)>>1;
  tmp = (PERIP(USB_RDPTR) + 1) & 0x03FF; 

  if (lenW <= (ENDPOINT_SIZE_1+1)>>1) {
    RingBufCopyX(packet->payload, (void *)(tmp + USB_RECV_MEM), lenW);
    PERIP(USB_RDPTR) = (lenW + tmp) & 0x03FF;
  }
  return ep & 0x7f;
}



u_int16 audioStash[91];
// Interrupt handler for USB perip
auto void Interrupt0() {
  register u_int16 g = PERIP(USB_STATUS);
  register u_int16 st;

  if (PERIP(USB_RDPTR) != PERIP(USB_WRPTR)) {
    st = USEX(PERIP(USB_RDPTR) + USB_RECV_MEM);
    if (((st >> 10) & 0x0f) == AUDIO_ISOC_OUT_EP) {
      GetPacket((USBPacket*)audioStash);
      if (AudioBufFree() > 45) {
	Enable();
	AudioPacketFromUSB(&audioStash[1], (audioStash[0]+1)/2);
	Disable();
      }
    } else {
      /* not audio packet */
    }
  }
  edgeTime2 = PERIP(TIMER_T1CNTL);
}



/*
  We replace the MassStorage() functionality.
  Uses ScsiTestUnitReady() to check the MMC/SD removal.
  Only appears as a mass storage device when MMC/SD is present.
 */
auto void MyMassStorage(void) {

  if (mmc.state == mmcNA || mmc.errors) {
      return;    // No MMC -> no USB MassStorage
  }

  puts("MyMassStorage");
  voltages[voltAnaUSB] = 31;
  voltages[voltAnaPlayer] = 31; //test only
  PowerSetVoltages(&voltages[voltCoreUSB]);
  BusyWait10();  // a 2.5ms-10ms delay to let the voltages change (rise)
  LoadCheck(NULL, 1); /* 4.0x clock required! */
  SetRate(44100U); /* Our USB-audio requires 44100Hz */
  InitUSB(USB_MASS_STORAGE);  
  PERIP(TIMER_CONFIG) = 0; // 48 MHz

  // USB loop
  while (mmc.state != mmcNA && mmc.errors == 0) {
    USBHandler();
    PERIP(USB_CONTROL) |= USB_STF_RX;
    ScsiTaskHandler();

    if (ticks) {
      ticks = 0;
      KeyScan();
      if (!(ReadGPIO() & 0x80)) {
	// Record high quality
	KeyEventHandler(ke_rec_sp);
      }
      if (ReadGPIO() & KEY_5) {
	// Player
	KeyEventHandler(ke_mmcplay);
      }


    }
        
    /* If no SOF in such a long time.. */
    if (USBWantsSuspend()) {
      if (USBIsDetached()) {
	break;
      }

#if 0      
      if (0) //Go to suspend  NOTE: DOES NOT WORK (maybe should increase time?)
      {
	register u_int16 io = PERIP(GPIO1_ODATA);
	PERIP(GPIO1_ODATA) = io & ~(LED1|LED2);      /* LEDs off */
	PERIP(SCI_STATUS) |= SCISTF_ANA_PDOWN;       /*fix suspend*/
	USBSuspend(0); /* no timeout */
	PERIP(GPIO1_ODATA) = io;      /* LEDs back to old states */
	PERIP(SCI_STATUS) &= ~SCISTF_ANA_PDOWN;
	/* with USB audio the LED settings are handled by user interface */
      }
#endif      
      USB.lastSofTime = ReadTimeCount();
    }
  }
    
  // Leaving USB
  hwSampleRate = 1; /* Forces FREQCTL update at next SetRate() */
  PERIP(SCI_STATUS) &= ~SCISTF_USB_PULLUP_ENA; /* Take us 'out of the bus' */
  PERIP(USB_CONFIG) = 0x8000U; /* Reset USB */
  map->Flush(map, 1);  
  /* Set player voltages. No need to wait for the voltages to drop. */
  PowerSetVoltages(&voltages[voltCorePlayer]);
}

void ResetBulkEndpoints(void);
void MyDecodeSetupPacket(){
  static u_int16 setupPacketCount = 0;
  setupPacketCount++;      

  //EP0 shalt not be transmitting now.
  USB.EPReady[0] = 1; //EP0 shalt be ready now.
    
  if ((s_int16)USB.pkt.payload[0] >= 0 &&
      (USB.pkt.payload[USB_SP_REQUEST] & 0xff) ==
      USB_REQUEST_SET_INTERFACE) {
    USBResetEndpoint(0);
    /* index = interface, value = alternate setting */
    USB.interfaces =
      (USB.pkt.payload[USB_SP_VALUE] & 0xff00U) |
      (USB.pkt.payload[USB_SP_INDEX] >> 8);
#if 1 /* ok with windoze */
    if ( USB.pkt.payload[USB_SP_INDEX] == 0x0300 /*0x0003 (MSC)*/){
      ResetBulkEndpoints();
    }
    if ( USB.pkt.payload[USB_SP_INDEX] == 0x0100 /*0x0001 (SPK)*/) {
      USBResetEndpoint(0x01);
    }
    if ( USB.pkt.payload[USB_SP_INDEX] == 0x0200 /*0x0002 (MIC)*/){
      /* Input Endpoint (device writes to PC) */
      USBResetEndpoint(0x81);
      PERIP(USB_EP_ST1) |= USB_STF_IN_ISO;
    }
#endif
    USBSendZeroLengthPacketToEndpoint0(); 
  }
  /// \todo: add handler for clearfeature:endpointhalt for epMicOrSomething (error@start/end of rec)
  else {    
    RealDecodeSetupPacket();
  }
}


// Call wrapper to mass storage request handler
void MyMSCPacketFromPC(USBPacket *inPacket){
  RealMSCPacketFromPC(&USB.pkt);
  PERIP(USB_EP_ST3) &= ~USB_STF_IN_FORCE_NACK;
}


u_int16 epPackets[4];
u_int32 bytesReceived=0;
u_int16 btimeCount=0;
// New USB handler
void MyUSBHandler(){
  __y u_int16 ep;

  if (PERIP(USB_STATUS) & 0x8000U) {
    InitUSB(0);
    PERIP(USB_EP_ST1) |= USB_STF_IN_ENABLE | USB_STF_IN_ISO;
    SCSI.DataOutBuffer = SCSI.DataBuffer;
    PERIP(USB_EP_ST3) &= ~USB_STF_IN_FORCE_NACK; // Release EP3
  }
  
  if (USB.EPReady[0] == 0) {
    USBContinueTransmission(0);
  }
  if (USB.EPReady[MSC_BULK_IN_ENDPOINT] == 0) {
    USBContinueTransmission(MSC_BULK_IN_ENDPOINT);
  }

  // Are there packet(s) in the receive buffer?
  if (PERIP(USB_RDPTR) != PERIP(USB_WRPTR)) {

    PERIP(USB_EP_ST3) |= USB_STF_IN_FORCE_NACK;
    ep = USBReceivePacket(&USB.pkt);
    
    if (USB.pkt.length) { //Does it actually contain data?
      if (ep == 0) { // Setup Packet
	DecodeSetupPacket();	
      } else if (ep == AUDIO_ISOC_OUT_EP) {
	AudioPacketFromUSB(USB.pkt.payload, (USB.pkt.length+1)/2);
      } else if (ep == MSC_BULK_OUT_ENDPOINT) {
	PERIP(INT_ENABLEL) |= INTF_USB; // Enable USB interrupt while MSC in progress
	MSCPacketFromPC(&USB.pkt);
	PERIP(INT_ENABLEL) &= ~INTF_USB;
      } else {
	// Unknown packet in receive buffer
      }
    } else {
      // Empty packet in receive buffer
    }
  } else {
    // No packets in receive buffer
  }
  PERIP(USB_EP_ST3) &= ~USB_STF_IN_FORCE_NACK;

  if (PERIP(USB_STATUS) & USB_STF_SOF) {
    PERIP(USB_STATUS) = USB_STF_SOF;
    USB.lastSofTime = ReadTimeCount();
    USB.lastSofFill = AudioBufFill();
  }
}




// Ok, Let's Go.
void main(void) {

  puts("mmcmi.c");
  
  voltages[voltIoPlayer] = 31; //test only
  voltages[voltAnaPlayer] = 31; //test only
  PowerSetVoltages(&voltages[voltCorePlayer]);

#if 1

  PERIP(GPIO0_MODE) = 0x0BFF;
  PERIP(GPIO0_DDR) = 0x3400;
  //PERIP(GPIO0_ODATA) = 0x0400;
  PERIP(GPIO0_ODATA) = 0x0400 | UR3_LED7;

  PERIP(GPIO1_MODE) = 0x0033;
  PERIP(GPIO1_DDR) = 0x000C;
  PERIP(GPIO1_ODATA) = 0x0004;

#endif

  clockX = player.maxClock;
  InitAudio();
  StartADC();

  currentKeyMap = usbRecorderKeyMap;
  SetHookFunction((u_int16)KeyEventHandler, USBRecorderKeyEventHandler);


  
#if 1
  //Enable amplifier
  PERIP(GPIO0_DDR) |= GPIO0_ALE;
  PERIP(GPIO0_SET_MASK) = GPIO0_ALE;
#endif

  // Drop from USB
  PERIP(SCI_STATUS) &= ~SCISTF_USB_PULLUP_ENA;
  PERIP(USB_CONFIG) = 0x8000U;
  
  PERIP(GPIO1_ODATA) |=  LED1|LED2;
  PERIP(GPIO1_DDR)   |=  LED1|LED2;   //Output
  PERIP(GPIO1_MODE)  &= ~(LED1|LED2);  //GPIO

  PERIP(GPIO0_DDR)   |=  REC_LED;   //Output
  PERIP(GPIO0_MODE)  &= ~(REC_LED);  //GPIO

  
  player.volumeOffset = -24;
  player.pauseOn = 0;
  
  keyOld = KEY_POWER;
  keyOldTime = -32767;

  // Set some hook functions...
  SetHookFunction((u_int16)IdleHook, NullHook); //Disable keyboard scanning
  SetHookFunction((u_int16)USBHandler, MyUSBHandler);
  SetHookFunction((u_int16)DecodeSetupPacket, MyDecodeSetupPacket);  
  SetHookFunction((u_int16)MSCPacketFromPC, MyMSCPacketFromPC);
  SetHookFunction((u_int16)InitUSBDescriptors, MyInitUSBDescriptors);

  //player.volume = 0;
  PlayerVolume();

  //puts("Mapper Create");
  // Create logical disk interface
#ifdef USE_MMC
  map = FsMapMmcCreate(NULL, 0);
  InitializeMmc(500);
#else 
  map = FsMapSpiFlashCreate(NULL, 0); 
#endif




  //puts("While1");
  //while(1){
  // When USB is attached, go to mass storage handler
  if(USBIsAttached()) {                 
    MyMassStorage();    
  }
  //}
  
  StopADC();
  SpiExec(STANDALONE_PLAYER_MODE);
  

}

