// file mmcr.c - MMC/SD card USB audio recorder

//#define puts ERROR_DO_NOT_USE_PUTS
#define puts(a);

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vs1000.h>
#include <mappertiny.h>
#include <minifat.h>
#include <usb.h>
#include <usblowlib.h>
#include <dev1000.h>
#include <codec.h>
#include <audio.h>
#include "scsi.h"
#include <player.h>
#include "multi.h"

#define USE_44K
#define DEBUG_LEVEL 0
#define USE_HC      /* Support SD-HC cards. (42 words) */
#define EXTEND_DIRECTORY
#define MMC_USE_MULTIWRITE

const /*__y*/ u_int16 myConfigurationDescriptor[] =
{
  0x0902, //Len | Type
  0x6400, //TotLenLo    | //TotLenHi,
  0x0201, //Interfaces  | //Configurations
  0x0080, //Config.Name | //Attr
  0x3209, //100mA       | //AudioControlInterface Length
  0x0400, //Type        | //Interface 0
  0x0000,  //Selected by Alternate Setting 0  |//Number of Endpoints Used
  0x0101,  //Class:Audio | //Subclass: Audio Control
  0x0000,  //Protocol    | //Interface name string
  
  //AudioControlClassInterface
  0x0924, //Length, Type = CS_INTERFACE
  0x0100,  //Subtype: Header
  0x011E, //Audio Class spec v 1.0  |//ClassLenLo  
  0x0001,  //ClassLenHi  |//number of audio stream interfaces
  0x010C, //First (and only) Streaming interface is 1
  
  //Input Terminal Descriptor
  0x2402, //Length, Type = CS_INTERFACE
  0x0101, //Terminal ID
  0x0100, //type = USB Streaming
      //Audio Cluster
  0x0200,  //n. of logical output channels
  0x0000, //Undefined Spatial Positions
  0x0009, //terminal name string

  //Output Terminal Descriptor
  0x2403, //Len, type = CS_INTERFACE
  0x0204, //Terminal ID
  0x0300, //Terminal type: Desktop Speaker
  0x0100, //ID of source terminal: USB Streaming terminal 1
  
  //Class Length is AudioControlClassInterface until here
  
  //AudioStreamingInterface Alternate Setting 0 (No sound)
  0x0904, //09=Length, 04=Type
  0x0100, //01 = Interface, 00=This describes Alternate Setting 0
  0x0001, //Number of interfaces used=0: only uses EP0, no sounddata
  0x0200, //01=Class:Audio, 02=Subclass: Audio Streaming
  0x0009, //00=Interface name string index. 09 belongs to next descr.
  
  //AudioStreamingInterface Alternate Setting 1 (Yes sound)
  0x0401, //09=Length, 04=Type, 01=Interface
  0x0101, //This is alternate setting 1 using 1 data interface
  0x0102, //Class: Audio, Subclass: Audio Streaming
  0x0000, //Protocol (unused, set to 0)
  
  //Audio Streaming Class Interface
  0x0724, //bLength=7, bDescriptorType=CS_INTERFACE(0x24)
  0x0101, //bDescriptorSubtype=AS_GENERAL(0x01), bTerminalLink=1
  AUDIO_DELAY_FRAMES*256+01, //bDelay=n*1ms, wFormatTag(lo 0x01)=PCM
  0x000B, //wFormatTag(hi 0x00)=PCM, next descriptor len 0x0b
  
  //Audio Format Type Descriptor
  0x2402, //Length=0x0b, Type=CS_INTERFACE(0x24), sub=FORMAT_TYPE(0x02)
  0x0102, //01=Type=FORMAT_TYPE_I (s_intNN PCM)  02=Number of channels
  0x0210, //02=Bytes in audio subframe (1, 2, 3 or 4).  10=16 bits/sample
  0x0180, //01=Number of supported frequencies: 1 discrete frequency
  0x3F00, //0x00AC44 = 44100 Hz
  
  //Endpoint Descriptor
  0x0905, //Length, Type
  0x0109, //Out Endpoint 1
  0xff03, //Endpoint Size 1023
  0x0100, //Polling Interval
  0x0007, //Synchronization info endpoint (unused)
  
  //ClassEndpointDescriptor
  0x2501, //Length, Type
  0x0000, //Attributes: None, Audio Path Delay Units: unused
  0x0000  //Audio Path Delay (unspecified)
};

#define MODEL_NAME_LENGTH 36
const u_int16 myModelNameStr[] = {
  ((MODEL_NAME_LENGTH * 2 + 2)<<8) | 0x03,
  'U'<<8,
  'S'<<8,
  'B'<<8,
  'R'<<8,
  'e'<<8, //5
  'c'<<8,
  'o'<<8,
  'r'<<8,
  'd'<<8,
  'e'<<8, //10
  'r'<<8,
  ' '<<8,
  'i'<<8,
  'n'<<8,
  ' '<<8, //15
  'L'<<8,
  'o'<<8,
  's'<<8,
  's'<<8,
  'l'<<8, //20
  'e'<<8,
  's'<<8,
  's'<<8,
  ' '<<8,
  'R'<<8, //25
  'e'<<8,
  'c'<<8,
  'o'<<8,
  'r'<<8,  //29
  'd'<<8, //30
  ' '<<8,
  'M'<<8,
  'o'<<8,
  'd'<<8,
  'e'<<8, //35
  '.'<<8
};




extern struct FsPhysical *ph;
extern struct FsMapper *map;
extern struct CodecServices cs;

// Amount of memory to use for file write buffering, excluding extra 45 words
// This must be multiple of 256 and 45 free words must follow

// NOTE: This version uses both X and Y memory for buffer, do not change!
#define BUFFER_SIZE (0x1800 + 0x2400)


u_int32 currentBlock;
u_int32 currentBlockFill;
//s_int16 __y *yBufStart = mallocAreaY;
//s_int16 __y *yBufEnd = mallocAreaY + Y_BUFFER_SIZE;
//s_int16 __y *yBufPtr;// = yBufStart;
//s_int16 __y *yBufRdPtr;// = yBufStart;
u_int16 interfaceStopped = 0;

u_int16 bufferSpace = BUFFER_SIZE;
u_int16 bufWi;
u_int16 bufRi;


void PUT_BUFFER(u_int16 a) {
  if (bufWi < 0x1800) {
    *(u_int16 *)bufWi = (a);
  } else {
    *(__y u_int16 *)(bufWi - 0x1800 + 0x1C00/*yOffset*/) = (a);
  }
  if (++bufWi >= BUFFER_SIZE) {
    bufWi = 0;
  }
}
 


#include "module_mmcrw.cc"

#define SOUND_DIR          "VOICE   "
#define SOUND_DIR_PACKED "\pVOICE   "

extern struct FsMapper *map;
#include "module_fatwrite.cc"

void FatPrepareForSaving(void) {
    u_int16 name[12/2];


#if 0
    static const u_int16 waveHeader[] = {
      /* 8000Hz 1-ch 16-bit signed WAV */
      0x5249,0x4646,0x2400,0x0000,0x5741,0x5645,0x666d,0x7420,
      0x1000,0x0000,0x0100,0x0100,0x401f,0x0000,0x803e,0x0000,
      0x0200,0x1000,0x6461,0x7461,0xffff,0xffff,0x0000,0x0000,
      0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000
      /* 10 zero-samples inserted to align to 32-word boundary */
    };       
#endif

#ifdef USE_44K
#if 1
    static const u_int16 waveHeader[] = {
      /* 44100Hz 2-ch 16-bit signed WAV */
      0x5249,0x4646,0xffff,0x0000,0x5741,0x5645,0x666d,0x7420,
      0x1000,0x0000,0x0100,0x0200,0x44ac,0x0000,0x10b1,0x0200,
      0x0400,0x1000,0x6461,0x7461,0xffff,0xffff,0x0000,0x0000,
      0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000
      /* 10 zero-samples inserted to align to 32-word boundary */
    };
#else
    static const u_int16 waveHeader[] = {
      /* 44100Hz 1-ch 16-bit signed WAV */
      0x5249,0x4646,0xffff,0x0000,0x5741,0x5645,0x666d,0x7420,
      0x1000,0x0000,0x0100,0x0100,0x44ac,0x0000,0x8858,0x0100,
      0x0400,0x1000,0x6461,0x7461,0xffff,0xffff,0x0000,0x0000,
      0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000
      /* 10 zero-samples inserted to align to 32-word boundary */
    };
#endif
#else
    static const u_int16 waveHeader[] = {
      /* 16000Hz 2-ch 16-bit signed WAV */
      0x5249,0x4646,0xffff,0x0000,0x5741,0x5645,0x666d,0x7420,
      0x1000,0x0000,0x0100,0x0200,0x803e,0x0000,0x00fa,0x0000,
      0x0400,0x1000,0x6461,0x7461,0xffff,0xffff,0x0000,0x0000,
      0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000
      /* 10 zero-samples inserted to align to 32-word boundary */
    };

#endif
    

    mmc.dirLine = -1;
    mmc.largestFreeSize = 0;
    //mmc.smallerFreeSize = 0;
    mmc.fileSize = 0;
    if (InitFileSystem()) {
	//mmc.dirBlock = 1234;
	return;
    }
    //mmc.fileSize = 0;
    if (!mmc.fileName[3]) {
	memcpyXY(mmc.fileName, "\pVOICE_00WAV", sizeof(mmc.fileName));
    }
    /* Must find free area first to be able to extend directory! */
    FatIterateOverFreeSectors(FsGetFreeSectors, NULL);
#if DEBUG_LEVEL > 0
    puthex(mmc.smallerFreeStart>>16);
    puthex(mmc.smallerFreeStart);
    puts("=smallerFreeStart");
    puthex(mmc.smallerFreeSize>>16);
    puthex(mmc.smallerFreeSize);
    puts("=smallerFreeSize");
    puthex(mmc.largestFreeStart>>16);
    puthex(mmc.largestFreeStart);
    puts("=largestFreeStart");
    puthex(mmc.largestFreeSize>>16);
    puthex(mmc.largestFreeSize);
    puts("=largestFreeSize");
    puthex((mmc.largestFreeStart - minifatInfo.dataStart) / minifatInfo.fatSectorsPerCluster);
    puts("=first data cluster");
#endif
    do {
	if ((++mmc.fileName[3] & 0xff) > '9') {
	    if (
//		((mmc.fileName[3] += 0x100-10) & 0xff00) > ('9'<<8)
		(mmc.fileName[3] += 0x100-10) > (('9'<<8)+0xff)
		) {
		mmc.fileName[3] = 0x3030; /* "00" */
		/* avoid '[' '\' ']' and '@' */
		if ((++mmc.fileName[2] & 0xff) > 'Z')
		    mmc.fileName[2] = (('E'<<8) + 'A');
	    }
	}
	memcpyYX(name, mmc.fileName, sizeof(name));
#if DEBUG_LEVEL > 1
	putchar(name[0]>>8);
	putchar(name[0]);
	putchar(name[1]>>8);
	putchar(name[1]);
	putchar(name[2]>>8);
	putchar(name[2]);
	putchar(name[3]>>8);
	putchar(name[3]);
	putchar('.');
	putchar(name[4]>>8);
	putchar(name[4]);
	putchar(name[5]>>8);
	puts("-Trying");
#endif
    } while (FatFindDirEntry(name) == FFDE_NAMEINVALID);
#if DEBUG_LEVEL > 1
    puts("-Got it");
#endif

  currentBlock = mmc.largestFreeStart;

  memcpy(0, waveHeader, sizeof(waveHeader));
  bufWi = 32;
  bufRi = 0;
  bufferSpace = BUFFER_SIZE-32;


}

u_int16 overPrint = 1;
u_int16 wordsMissed = 0;

/** USB (Interrupt) Handler */
void MyUSBHandler() {
  __y u_int16 ep;
  register u_int16 info;
  
  if (PERIP(USB_STATUS) & 0x8000U) {
#if DEBUG_LEVEL > 0
    puts("BRST");
#endif
    InitUSB(0);
  }
  if (PERIP(USB_STATUS) & USB_STF_SOF) {
    PERIP(USB_STATUS) = USB_STF_SOF; /** clear SOF */
    //USB.lastSofTimeout = 0;
    USB.lastSofTime = ReadTimeCount();
    USB.lastSofFill = AudioBufFill();
  }
  if (USB.EPReady[0] == 0) {
    USBContinueTransmission(0);
  }
  // Are there packet(s) in the receive buffer?
  if (PERIP(USB_RDPTR) != PERIP(USB_WRPTR)) {
    USBCheckForSetupPacket();

    info = USEX(PERIP(USB_RDPTR) + USB_RECV_MEM);
    ep = USBReceivePacket(&USB.pkt);
    
    if (USB.pkt.length) { //Does it actually contain data?
      if (info & 0x4000U) { // Setup Packet
        DecodeSetupPacket();
      } else if (ep == AUDIO_ISOC_OUT_EP) {

	if (bufferSpace > USB.pkt.length / 2){
	  register u_int16 i;
	  bufferSpace -= USB.pkt.length/2;
	  for (i=0; i < USB.pkt.length/2; i++) {
	    PUT_BUFFER(USB.pkt.payload[i]);
	  }
	}
	
#if 0
#if 0
	  bufferSpace -= USB.pkt.length/2;
	  memcpyXY(yBufPtr, USB.pkt.payload, USB.pkt.length / 2);
	  yBufPtr += USB.pkt.length / 2;
#else
	  bufferSpace -= USB.pkt.length/4;
	  {
	    register s_int32 v;
	    register u_int16 *s = USB.pkt.payload;
	    register u_int16 l = USB.pkt.length / 4;
	    while(l--){
	      v = (s_int16)SwapWord(*s++);
	      v += (s_int16)SwapWord(*s++);
	      v /= 2;
	      *yBufPtr++ = SwapWord((u_int32)v & 0xffffU);
	    }
	  }
#endif
	  if (yBufPtr > yBufEnd) {
	    register u_int16 __y *p = yBufEnd;
	    register u_int16 __y *d = yBufStart;
	    while (p != yBufPtr) {
	      *d++ = *p++;
	    }
	    yBufPtr = d;	    	     	  
	  }	  
	} 
#endif

	else {
	  wordsMissed += USB.pkt.length / 2;
	}
	AudioPacketFromUSB(USB.pkt.payload, (USB.pkt.length+1)/2);
      } else {
	//unknown packet
      } 
    } else {
      // zero-length packet
    }
  }
}
void MyDecodeSetupPacket(){
  USB.EPReady[0] = 1; //EP0 shalt be ready now.
  
  if (((s_int16)USB.pkt.payload[0] >= 0 ) &&
      ((USB.pkt.payload[USB_SP_REQUEST] & 0xff) ==
       USB_REQUEST_SET_INTERFACE) &&
      (USB.pkt.payload[USB_SP_VALUE] == 0))
    {
      interfaceStopped = 1;
    };
  RealDecodeSetupPacket();
}

void MyInitUSBDescriptors(u_int16 initDescriptors){
  RealInitUSBDescriptors(2); // ROM set descriptors for audio;
  USB.descriptorTable[DT_MODEL]  = myModelNameStr;
#ifndef USE_44K
  USB.descriptorTable[DT_CONFIGURATION] = myConfigurationDescriptor;
#endif
}

void SpiExec(u_int16 programOffset){
  ResetIRAMVectors();
  currentKeyMap = sixKeyMap; //in ROM
  PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1;
  PERIP(SPI0_FSYNC) = 0;
  PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (12-1);
  PERIP(INT_ENABLEL) = 0;
  PERIP(GPIO1_MODE) = 0x3f;	
  SpiLoad(programOffset,SPIADDR_EXTRABYTES);	    	  
}



void main() {
  puts("mmcr.c");
  clockX = player.maxClock;
  Enable();

  while(PERIP(UART_STATUS)&UART_ST_TXRUNNING)
    ;
  


  InitAudio(); /* goto 3.0x..4.0x */
  PERIP(INT_ENABLEL) = INTF_RX | INTF_TIM0;
  PERIP(INT_ENABLEH) = INTF_DAC;

  voltages[voltCoreUSB] = 29;
  voltages[voltIoUSB] = 31;
  PowerSetVoltages(&voltages[voltCoreUSB]);

  PERIP(GPIO1_ODATA) &= ~(LED1|LED2);      /* POWER led on */
  PERIP(GPIO1_DDR)   |=  LED1|LED2; /* SI and SO to outputs */
  PERIP(GPIO1_MODE)  &= ~(LED1|LED2); /* SI and SO to GPIO */
  PERIP(GPIO0_ODATA) &= ~(1 << 10); // NFCE low
  LED_ON(UR3_LED4);

  REC_LED_INIT;

  
  LoadCheck(NULL, 1); /* 4.0x clock required! */

  SetRate(44100U); /* Our USB-audio requires 44100Hz */

  map = FsMapMmcCreate(NULL, 0); /* Create MMC Mapper */

  FsInitializeMmc(50);  

  FatPrepareForSaving();
  mmcDoMultiWrite = 1;


  SetHookFunction((u_int16)IdleHook, NullHook); //Disable keyboard scanning
  SetHookFunction((u_int16)DecodeSetupPacket, MyDecodeSetupPacket); 
  SetHookFunction((u_int16)InitUSBDescriptors, MyInitUSBDescriptors);  

  InitUSB(2);


  bufferSpace = BUFFER_SIZE;

  interfaceStopped = 0;
  while (1) { 
    
    MyUSBHandler();  
    
    //    if (USBWantsSuspend()) {
    //      interfaceStopped = 2;
    //    }
    
    if (mmc.state == mmcWriteWait) {
      register u_int16 c;
#if 0
      while (SpiSendReceiveMmc(0xff00U,8) != 0x00ffU){
	//wait
      }
      mmc.state = mmcOk; /*write finished*/
      REC_LED_OFF;
#else
      if (c=SpiSendReceiveMmc(0xff00U, 8) == 0x00ffU) {
	mmc.state = mmcOk; /*write finished*/
	REC_LED_OFF;
#if DEBUG_LEVEL > 1
	if (wordsMissed > 400) {
	  puthex(wordsMissed); 
	  puts("=wordsMissed");
	  wordsMissed = 0;
	}
#endif
      }
#endif
    } else {


      if (bufferSpace < (BUFFER_SIZE - 255)) {
	if (bufRi < 0x1800) {
	  memcpy(minifatBuffer, (u_int16 *)bufRi, 256);
	}else{
	  memcpyYX(minifatBuffer, (__y u_int16 *)(bufRi-0x1800+0x1c00), 256);
	}	
	bufferSpace += 256;
	REC_LED_ON;
	FsMapMmcWrite(NULL /*no wait*/, currentBlock, 1, minifatBuffer);
	mmc.fileSize += 512;
	if (mmc.fileSize / 512 >= mmc.largestFreeSize) {
	  // End of disk space (end of largest fragment)
	  puts("EOD");
	  interfaceStopped = 2;	  
	}	  
	currentBlock++;	  

	bufRi += 256;
	if (bufRi >= BUFFER_SIZE) {
	  bufRi = 0;
	}	
      } 

      if (PERIP(SCI_STATUS) & SCISTF_REGU_POWERBUT) {
	// PC/STOP button pressed
	puts("PWRBUT");
	interfaceStopped = 2;
      }

      if (ReadGPIO() & KEY_5) {
	// PC/STOP button pressed
	puts("BUT 5");
	interfaceStopped = 3;
      }

      {
	static u_int16 myTimeCount;
	if (((++myTimeCount >> 6) & 0x0f) == 0) {
	  REC_LED_ON;
	} else {
	  REC_LED_OFF;
	}
      }

      if (interfaceStopped){
	if (mmc.fileSize > 2048) { //some arbitrary non-zero min file size
	  //puts("Write File");

	  FsFatMakeFile();
	  FatPrepareForSaving();
	}
	if (interfaceStopped > 1){
	  //puts("Exit");
	  Disable();
	  PERIP(GPIO1_MODE) = 0x3f;
	  if (interfaceStopped == 2){
	    SpiExec(PC_PLAYER_MODE);	
	  }else{
	    SpiExec(STANDALONE_PLAYER_MODE);
	  }
	}
	interfaceStopped = 0;
      }
    }
  }
}
