u_int32 originalDacInt;
u_int32 originalUsbInt;
u_int32 originalGpio0Int;

 /* Microphone Buffer Ptr */
/* Use USB out-packet-buffer 6 for storing sample data */
s_int16 *micBufStart = USB_SEND_MEM + (6 * 64);
s_int16 *micBufEnd = USB_SEND_MEM + (6 * 64) + 32;
  s_int16 *micBufPtr = USB_SEND_MEM + (6 * 64);
volatile u_int16 edgeTime = 0;
volatile u_int16 edgeTime2 = 0;

void StartADC(){
  //Save original int vector
  originalDacInt = ReadIRam(0x20+INTV_DAC);
#if USE_USB_INTERRUPT
  originalUsbInt = ReadIRam(0x20+INTV_USB);
#endif
  originalGpio0Int = ReadIRam(0x20+INTV_GPIO0);

  PERIP(INT_ENABLE) = 0; // Disable All Interrupts
  //Set new int vector
  SetInterruptVector_GPIO0();
  SetInterruptVector_DAC();
#if USE_USB_INTERRUPT
  WriteIRam (0x20+INTV_USB, ReadIRam((u_int16)InterruptStub0));
#endif

  //Mic input GPIOs
  PERIP(GPIO0_CLEAR_MASK) = (1<<MIC_RESET_BIT); // output level is low
  PERIP(GPIO0_MODE) &= ~(1<<MIC_RESET_BIT); //is GPIO
  PERIP(GPIO0_DDR) &= ~(1<<MIC_INPUT_BIT);; //is input
  PERIP(GPIO0_MODE) &= ~(1<<MIC_INPUT_BIT); //is GPIO
  PERIP(GPIO0_INT_FALL) |= (1 << MIC_INPUT_BIT); //Falling edge int

  //Configure Timer1 for ramp measurement
  PERIP(TIMER_T1L) = 30000; //374=16 kHz //2176 ~is about 11025
  PERIP(TIMER_T1H) = 0x0;
  PERIP(TIMER_T1CNTL) = 1000; //Current Value Low
  PERIP(TIMER_T1CNTH) = 0; //Current Value High
  PERIP(TIMER_CONFIG) = 0; //Timer Clock is 48MHz
  PERIP(TIMER_ENABLE) |= (1 << 1); //Enable timer 1

  // Enable Interrupts
  Enable();
  PERIP(INT_ENABLEH) = 0; 
  PERIP(INT_ENABLEL) = INTF_GPIO0 | INTF_RX | INTF_DAC;

}

void StopADC(){

  Disable();	
  PERIP(INT_ENABLEH) = 0; 
  PERIP(INT_ENABLEL) = INTF_RX | INTF_DAC;
  PERIP(TIMER_ENABLE) &= ~(1 << 1); //Disable timer 1
  WriteIRam (0x20+INTV_DAC, originalDacInt);
#if USE_USB_INTERRUPT
  WriteIRam (0x20+INTV_USB, originalUsbInt);
#endif
  WriteIRam (0x20+INTV_GPIO0, originalGpio0Int);
  Enable();

}

void InterruptHandler_Timer1(void){}

void InterruptHandler_DAC(void){ 
  static u_int16 mode = GOING_UP;
  
  if (mode == GOING_UP) {
    static s_int16 currentSample;
    static s_int16 previousSample;
    

    PERIP(TIMER_CONFIG) = 0; // 48 MHz timer stepping rate; max@44.1 = 1088 steps
    mode = RESETTING;
    PERIP(GPIO0_ODATA) &= ~(1<<MIC_RESET_BIT); //output state is low
    PERIP(GPIO0_DDR) |= (1<<MIC_RESET_BIT); //is output
    
    // If edgeTime is sane (set at GPIO0 interrupt), use as sample
    previousSample = currentSample;
    if (edgeTime > MIC_TIMER_START-1040 && edgeTime < MIC_TIMER_START-40) {
      currentSample = ((s_int16)((edgeTime)-MIC_OFFSET) * (MIC_GAIN/2));
    }    
    *micBufPtr++ = SwapWord(currentSample+(previousSample/2));
    if (micBufPtr > micBufEnd) micBufPtr = micBufStart;

    // EP1 IN is isochronous. Make it so. Always.
    PERIP(USB_EP_ST1) |= USB_STF_IN_ENABLE | USB_STF_IN_ISO;
    
    // If PC wants mic data, put it to xmit buffer
    if (PERIP(USB_EP_ST1)& USB_STF_IN_EMPTY){
      register s_int16 *i;
      register s_int16 *p;
      register u_int16 n;
      p = (s_int16 *) USB_SEND_MEM + (5 * 64); // use send buffer 3
      i = micBufStart;
      n = 0;
      while (i < micBufPtr){
	*p++ = *i++;
	n++;
      } 
      micBufPtr = micBufStart;
      PERIP(USB_EP_SEND1) = 0x8000 | (5 << 10) |(n*2);
    }

           
  }else{

    //PERIP(INT_ENABLEL) &= ~INTF_USB;
    mode = GOING_UP;

    PERIP(GPIO0_DDR) &= ~(1<<MIC_RESET_BIT); //is input (HiZ)
    PERIP(TIMER_T1CNTL) = MIC_TIMER_START; //Current Value Low

    /* Handle timeCount here, because timer interrupt is disabled.
    USBWantsSuspend() does not work without timeCount. NOTE:SUSPEND DOES NOT WORK*/
    /*22050Hz*/
    {
      static int cnt = 0;
      if (++cnt > 22050/TIMER_TICKS) {
	cnt = 0;
	timeCount++;
      }
    }

  }
}


