#include <romfont.h>
struct screen {
    u_int16 X, Y, invertMask;
} screen;
const unsigned char oledInit[]={
  0xae, // Display off
  0xd3, // Display offset
  0x00, // 'value
  0xa8, // Multiplex ratio
  0x3f, // 'value
  0xad, // DC/DC on
  0x8b, // 'value 8b
  0x40, // Display start line
  0xa0, // Segment remap
  0xc8, // Scan direction
  0x81, // Set Contrast (=brightness)
  0x30, // 00 is minimum, ff is maximum
  0xb0, // Page address
  0x00, // Column address low nibble
  0x10, // Column address high nibble
  0xa4, // Entire display on
  0xa6, // Normal (not inverted) display
  0xaf, // Display on
  0xd8, // Power Save
  0x05  // Low power mode
};
#define SPI_MASTER_8BIT_CSHI    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN8 | SPI_CF_FSIDLE1
#define SCREEN_SET_COMMAND_MODE PERIP(GPIO1_CLEAR_MASK) = 0x0004
#define SCREEN_SET_DATA_MODE    PERIP(GPIO1_SET_MASK) = 0x0004
#define SCREEN_SELECT           PERIP(GPIO0_CLEAR_MASK) = 0x4000
#define SCREEN_END              PERIP(GPIO0_SET_MASK) = 0x4000 


// Write command c to display controller
void ScreenPutCommand(register __a0 unsigned char c) {
  SCREEN_SET_COMMAND_MODE;
  SCREEN_SELECT;
  //SPI_MASTER_8BIT_CSHI; // SPI's "own" CS remains high  
  SpiSendReceive(c);
  SCREEN_END;
}

// Set x,y pointer of display controller
void ScreenLocate(register __b0 int x, register __b1 int y){
  screen.X = x;
  screen.Y = y;
  ScreenPutCommand(y | 0xb0);
  ScreenPutCommand(x & 0x0f);
  ScreenPutCommand((x >> 4) | 0x0010);
  //SCREEN_END;
}


// Write image data octet to display controller
void ScreenPutData(register __a1 u_int16 c) {
  SCREEN_SET_DATA_MODE;
  SCREEN_SELECT;
  //SPI_MASTER_8BIT_CSHI; // SPI's "own" CS remains high
  SpiSendReceive(c ^ screen.invertMask);
  SCREEN_END;

  SCREEN_SET_COMMAND_MODE;
  SCREEN_SELECT;
  //SPI_MASTER_8BIT_CSHI; // SPI's "own" CS remains high  
  SpiSendReceive(0xE3);
  SCREEN_END;
  ++screen.X;
  if (screen.X > 131) {
    ScreenLocate(0, ++screen.Y);
  }
}
void ScreenInit(){
  PERIP(SPI0_FSYNC) = -1; // keep SPI CS high during transfer!
  SPI_MASTER_8BIT_CSHI;
  PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (1-1); // Spi clock mclk/4 = max 12 MHz
  PERIP(GPIO1_MODE) |= 0x1f; /* enable SPI pins */

  PERIP(GPIO0_MODE) &= ~0x4000; //CS2 is GPIO (Display Chip Select)
  PERIP(GPIO0_DDR)  |= 0x4000; //CS2 is Output

  PERIP(GPIO1_MODE) &= ~0x0004; //SI is GPIO (Display Command/Data)
  PERIP(GPIO1_DDR)  |= 0x0004; //SI is Output

  SCREEN_SET_COMMAND_MODE;
  SCREEN_SELECT;
  {
      register const unsigned char *c = oledInit;
      register int i;
      //SPI_MASTER_8BIT_CSHI; // SPI's "own" CS remains high
      for (i=0; i<sizeof(oledInit); i++)
	  SpiSendReceive(*c++);
  }
  SCREEN_END;
  screen.X = screen.Y = screen.invertMask = 0;
}

// Draw character c to screen using custom data output function
// This function is responsible for character mapping.
void ScreenPutChar(register __c0 u_int16 c) {
  register u_int16 length;
  register u_int16 bit_index;
  register const u_int16 *p;
  
  if (c < 32 || c > 210) { //Invalid char
    return;
  }
  if (c > 132) { //Fixed width charset
    length = 5;    
    bit_index = 8;
    p = &fontData[5*(c-133)];
  } else { //variable width charset ("normal" ascii)
    // ASCII ' ' (32 decimal) is in ROM index 0
    length = fontPtrs[c-31] - fontPtrs[c-32];
    bit_index = 0;
    p = &fontData[fontPtrs[c-32]];
  }
  /* prevent wrap in the middle of character */
  if (screen.X + length > 132) {
    ScreenLocate(0, ++screen.Y);
  }
  {
      register u_int16 i;
      for (i=0; i<length; i++){
	  ScreenPutData((s_int16)(*p++) >> bit_index);
      }
  }  
  ScreenPutData(0);
}



// Screen Clear
void ScreenClear(){
    register u_int16 x, y;
    for (y=0; y<8; y++){
	ScreenLocate(0, y);
	for (x=0; x<132; x++){
	    ScreenPutData(0);
	}
    }    
    ScreenLocate(0, 0);
}


// Put packed string in Y ram (mainly for file name output)
void ScreenPutPacked(__y const u_int16 *d, u_int16 maxChars, u_int16 xEnd) {
    register int i;
    for (i=0; i<maxChars; i++) {
	register u_int16 c = *d++;
	if ((c & 0xff00U) == 0)
	    break;
	ScreenPutChar(c>>8);
	if ((c & 0xffU) == 0)
	    break;
	ScreenPutChar(c & 0xff);
    }
    i = screen.X;
    for (; i < xEnd; i++) {
	ScreenPutData(0);
    }            
}
void ScreenPutPackedX(const u_int16 *d, u_int16 maxChars, u_int16 xEnd) {
    register int i;
    for (i=0; i<maxChars; i++) {
	register u_int16 c = *d++;
	if ((c & 0xff00U) == 0)
	    break;
	ScreenPutChar(c>>8);
	if ((c & 0xffU) == 0)
	    break;
	ScreenPutChar(c & 0xff);
    }
    i = screen.X;
    for (; i < xEnd; i++) {
	ScreenPutData(0);
    }            
}

// Draw a u_int16 value to the screen as decimal number
void ScreenPutUInt(register u_int32 value, register u_int16 length){
  register u_int32 i; 
  u_int16 s[8];

  for (i=0; i<8; i++){
    s[i]=value % 10;
    value = value / 10;
  }
  do {
    ScreenPutChar(SYMBOL_NUMBER + s[--length]);
  } while (length);     
}
